#! /usr/bin/tclsh8.4

# src/sourceinstall.tcl.  Generated from sourceinstall.tcl.in by configure.

# Source Installer v. 0.4 Copyright (C) 2005 Claudio Fontana
#
# This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#     along with this program (look for the file called COPYING);
#     if not, write to the Free Software Foundation, Inc.,
#         51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#     You can contact the author (Claudio Fontana) by sending a mail
#     to sick_soul@users.sourceforge.net
#
# Per Gunde:
#     buon compleanno!



# the list of all packages
set packages {}
# the list of supported source compression formats
set src_compress_formats [ list "bz2" "gz" "Z" ]
# the array of all widgets
array set w {}

# global expect variables:
# timeout
# It does not work to simply set a global timeout -1 here, because
# package require Expect has the effect of overwriting the value.
set timeout -1
# the current spawned process spawn_id
set spawn_id 0


# how many times a SIGINT has been detected. Reset by lock_actions.
set sigint_seen 0
# whether actions are locked
set actions_locked 0
# use as return value from dialogs
set dialog_retval ""
# use as array return value from dialogs
array set dialog_retarr {}
# the about text (see below)
set about ""

proc new_package_info {} {
    return [list \
		selected no          \
		name ""              \
		description ""       \
		long_description ""  \
		configured no        \
		source_available no  \
		source_size ""       \
		source_location ""   \
		build_subdir ""      \
		installed no         \
		installed_size ""    \
		installed_files [list]]
}    

# the array of information about current selected source package
array set package_info [ new_package_info ]

proc new_options {} {
    return [list \
		prefix "/usr/local"     \
		manual_configure 1      \
		install 1               \
		keep_source 1           \
	        strip 0                 \
		src_compress "bz2"      \
	        add_directory ""]
}

# the array of options
array set o [ new_options ]

proc new_tips {} {
    return [list \
		prefix "This is the default prefix to use when installing software packages. Programs and data will generally be installed in a subtree of the specified directory. The initial value is '/usr/local' and is a good default for system-wide installs." \
		manual_configure "If this option is active you will be able to see the software configuration window. It will get you aquainted with common options supported by the packages, so it is suggested to keep this selected." \
		install "If this option is active, then adding a package will install the software. In most cases you want this active, but you might want to add the source only (see the keep source option), and perform installation only later." \
		keep_source "If this option is active, then adding a package will add the configured compressed source too. In most cases you want this active (this permits cleaner uninstallation among other advantages). However, if you know what you are doing and want to save some disk space you might wish to disable it." \
		strip "This option is initially disabled and it is recommended that it remains so. Enabling it will enable stripping of binaries prior to installation." \
		src_compress "This option controls the compression format to use when archiving source packages (see keep source option). Initially it is set to .bz2 (a very space efficient algorithm), but another reasonable value is .gz (less space efficient but much faster)" \
		add_directory "This option controls the default directory to open when browsing for packages to add. If left empty (the initial value), then the current working directory is used. If you tend to add packages from the same directory over and over again you can specify your alternative here. For example, /home/user/downloads" \
		package_name "This is the name of the package, as it appears in the package list. When modifying the name, some dangerous characters will not be allowed." \
		package_description "This is a short description of the package (single line)" \
		package_long_description "This description follows the short package description, and can span multiple lines." \
		package_name_auto "Assign the default value for the package name, which is the name of the top source directory in the source package." \
		package_name_manual "Specify a package name manually in the adjacent text entry" \
		add_choose_file "When this button is active, the Browse button on the right is used to choose a package in the form of a file release." \
		add_choose_dir "When this button is active, the Browse button on the right is used to choose a package in the form of a source directory.\nThis could be necessary if you wish to add code fetched using a versioning system (CVS, svn and similar)."  \
	   ]
}

# popup tips
array set tips [ new_tips ]

# our directories
set rundir [pwd]
set prefix [file dirname [file dirname [file normalize $argv0]]]
set cfgdir "$env(HOME)/.sourceinstall"
set packdir "$cfgdir/packages"
set srcdir "$cfgdir/src"
set builddir "$cfgdir/build"
set destdir "$cfgdir/install-destdir"
# set icondir "$prefix/share/sourceinstall/icons"

# files
set cfgfile "$cfgdir/sourceinstallrc"
set lockfile "$cfgdir/.sourceinstall_lock"

# icons

## image create photo image_folder_closed -file "$icondir/folder_closed.gif"
## image create photo image_folder_open -file "$icondir/folder_open.gif"

proc src_assert { assertion_string value } {
    if { $value == 0 } {
	src_error "Assertion failure" "assertion failed:\n\n$assertion_string\n\n"
	clean_exit 1
    }
}

proc src_error { title text } {
    dialog_modal_msg "Error" error $title $text
}

proc src_warning { title text } {
    dialog_modal_msg "Warning" warning $title $text
}

proc src_info { title text } {
    dialog_modal_info "Information" info $title $text
}

proc src_message { title text } {
    dialog_modal_msg $title info $title $text
}

# the real file type of FILENAME. A link will report "link".
proc file_ltype { filename } {

    if { [catch { file type $filename } t] } {
	return "none"
    }

    return $t
}

# the file type of FILENAME after all links have been resolved.

proc file_type { filename } {

    if { [catch { file stat $filename struct_stat }] } {
	return "none"
    }

    return $struct_stat(type)
}

proc is_valid_filestring { filestring } {
    return [expr ![regexp {[\x00-\x1f/;&()*?<>|=@ ]} $filestring]]
}

# this is a replacement for tk_dialog (complete rewrite), since
# tk_dialog is bugged. tk_dialog calls grab on its window, and this causes
# exception if the user is currently dragging the scrollbar.

proc src_dialog { title text bitmap default args } {
    global w
    global dialog_retval

    set dialog_retval -1

    if { ![string is integer $default] } { return -1 }
    if { $default >= [llength $args] } { return -1 }

    set w(modal_dialog) [toplevel .modal_dialog]
    wm title $w(modal_dialog) $title

    set w(frame_dialog_top) [frame $w(modal_dialog).top -padx 10 -pady 10]
    set w(frame_dialog_bottom) [frame $w(modal_dialog).bottom -padx 10 -pady 10]

    pack $w(frame_dialog_top) -side top -fill both -expand 1
    pack $w(frame_dialog_bottom) -side top

    set w(label_dialog_text) [label $w(frame_dialog_top).text -justify left -text $text -padx 12 -pady 24 -font font_text -foreground midnightblue -background antiquewhite -relief sunken -bd 1 -wraplength 400]

    pack $w(label_dialog_text) -side right -expand 1 -fill both -padx 10

    if { $bitmap != "" } {
	set w(label_dialog_bitmap) [label $w(frame_dialog_top).bitmap -bitmap $bitmap]
	pack $w(label_dialog_bitmap) -side left -padx 8
    }

    set i 0
    foreach this $args {
	set w("button${i}_dialog") [button "$w(frame_dialog_bottom).button$i" -text $this -width 5]
	$w("button${i}_dialog") configure -command "
            set dialog_retval $i
	    destroy $w(modal_dialog)"

	pack $w("button${i}_dialog") -side left -padx 10
	incr i
    }

    focus $w("button${default}_dialog")

    dialog_wait $w(modal_dialog) 0
    return $dialog_retval
}


# this procedure is called by modal dialogs to grab input and wait until
# the dialog dies

proc dialog_wait { modal_window center } {
    wm resizable $modal_window 0 0
    wm transient $modal_window .
    wm withdraw $modal_window
    update idletasks

    if { $center } {
	set x [expr ([winfo screenwidth $modal_window] - [winfo reqwidth $modal_window]) / 2]
	set y [expr ([winfo screenheight $modal_window] - [winfo reqheight $modal_window]) / 2]

    } else {
	set x [expr [winfo rootx .]+50]
	set y [expr [winfo rooty .]+50]
    }

    wm geometry $modal_window "+$x+$y"
    wm deiconify $modal_window

    tkwait visibility $modal_window
    focus $modal_window

#   bind $modal_window <KeyPress-Return> \
	"event generate $modal_window <KeyPress-space>"
#   bind $modal_window <KeyRelease-Return> \
	"event generate $modal_window <KeyRelease-space>"

    grab $modal_window
    tkwait window $modal_window
}

# the following procedure assigns a popup tip to a widget
# note that widthw is a widget to get the desired width from.

proc tip { widget widthw text } {

    bind $widget <Enter> "after 1000 { tip_show %W $widthw { $text } }"
    bind $widget <Leave> "catch { destroy .modal_tip }
                        after cancel { tip_show %W $widthw { $text } }"
}

proc tip_show { args } {
    global w

    set widget [lindex $args 0]
    set widthw [lindex $args 1]

    if { ![winfo exists $widget] } { return }

    if { [llength $args] > 2 } {
	set text [lindex $args 2]
    } else {
	set text "No description available."
    }

    catch { destroy .modal_tip }
    toplevel .modal_tip
    wm overrideredirect .modal_tip 1

    set x [expr [winfo rootx $widget]+[winfo width $widget]/4]
    set y [expr [winfo rooty $widget]+[winfo height $widget]]

    # see? the width is calculated from the widthw parameter
    set width [winfo width $widthw]

    frame .modal_tip.frame -bd 1 -relief groove -background \#ffffc0
    label .modal_tip.frame.label -text $text -bd 0 -background \#ffffc0 -foreground midnightblue -wraplength $width -justify left -padx 8 -pady 4

    pack .modal_tip.frame
    pack .modal_tip.frame.label -anchor w -side left
    wm geometry .modal_tip "+$x+$y"

    .modal_tip.frame configure -height [winfo reqheight .modal_tip.frame.label] -width [expr $width+16]
    pack propagate .modal_tip.frame 0
}

proc set_default_size {} {
    set width [winfo screenwidth .]
    set height [winfo screenheight .]

    set width [expr $width*6/9]
    set height [expr $height/2]

    wm minsize . $width $height
    wm geometry . ${width}x${height}
}

proc build_menu {} {
    global w
    global about

    set w(main_menu) [menu .main_menu -font font_text]
    . configure -menu $w(main_menu)

    set w(menu_file) [menu $w(main_menu).menu_file]
    set w(menu_edit) [menu $w(main_menu).menu_edit]
    set w(menu_window) [menu $w(main_menu).menu_window]
    set w(menu_help) [menu $w(main_menu).menu_help]

    $w(main_menu) add cascade -label "File" -menu $w(menu_file) -underline 0
    $w(main_menu) add cascade -label "Edit" -menu $w(menu_edit) -underline 0
    $w(main_menu) add cascade -label "Window" -menu $w(menu_window) -underline 0
    $w(main_menu) add cascade -label "Help" -menu $w(menu_help) -underline 0

    $w(menu_file) add command -label "Quit" -underline 0 -accelerator "CTRL+Q" -command { clean_exit 0 }


    $w(menu_edit) add command -label "Cut" -underline 2 -accelerator "CTRL+X" -state disabled
    $w(menu_edit) add command -label "Copy" -underline 0 -accelerator "CTRL+C" -state disabled
    $w(menu_edit) add command -label "Paste" -underline 0 -accelerator "CTRL+V" -state disabled

    $w(menu_edit) add separator
    $w(menu_edit) add command -label "Edit package info" -underline 0 -state disabled -accelerator "E" -command {
	dialog_modal_package_info
    }

    $w(menu_edit) add separator
    $w(menu_edit) add command -label "Preferences..." -underline 0 -accelerator "CTRL+O" -command {
	dialog_modal_preferences
    }

    $w(menu_window) add command -label "Fullscreen" -underline 0 -state disabled
    $w(menu_window) add command -label "Default size" -underline 0 -command {
	set_default_size
    }
    $w(menu_window) add checkbutton -label "Save size and position on exit" -underline 0 -onvalue "yes" -offvalue "no" -state disabled

    $w(menu_help) add command -label "Help contents" -underline 0 -accelerator "F1" -state disabled
    $w(menu_help) add separator
    $w(menu_help) add command -label "About Source Installer" -underline 0 -command {
	src_info "Source Installer version 0.4" $about
    }
}

proc init_wm {} {
    global packages
    global w

    wm protocol . WM_DELETE_WINDOW { clean_exit 2 }

    set_default_size
    wm title . "Source Installer"

    build_menu

    # main
    set w(frame_main) [frame .main]

    # pane_vert
    set w(pane_vert) [panedwindow $w(frame_main).pane_vert -showhandle 1 -handlepad 100 -sashpad 5]

    # frame_console
    set w(frame_console) [frame $w(pane_vert).frame_console -padx 1]

    # label_console
    set w(label_console) [label $w(frame_console).label_console -text "Information" -font font_title -justify center -foreground "darkred"]

    # console
    set w(console) [text $w(frame_console).console -padx 10 -pady 10]

    $w(console) tag configure tag_title -font font_subtitle -justify center -foreground "darkred"
    $w(console) tag configure tag_subtitle -font font_text -justify left -foreground "darkred"
    $w(console) tag configure tag_code -font font_code -justify left -wrap char -foreground "black"
    $w(console) tag configure tag_command -font font_command -justify left -wrap char -foreground "black"
    $w(console) tag configure tag_comment -font font_comment -justify left -wrap char -foreground "darkred"
    $w(console) tag configure tag_err -font font_subtitle -justify left -foreground "black"

    # scroll_console
    set w(scroll_console) [scrollbar $w(frame_console).scroll -command "$w(console) yview" ]

    $w(console) configure -yscrollcommand "$w(scroll_console) set"

    # frame_packages
    set w(frame_packages) [frame $w(pane_vert).frame_packages]

    # list_packages
    set w(list_packages) [listbox $w(frame_packages).list_packages -listvariable packages -exportselection 0]

    # popup menu associated with package list
    set w(popup_action) [menu .popup_action -font font_text]
    $w(popup_action) add command -label "Remove" -underline 0 -accelerator "R" -state normal -command "action_remove"
    $w(popup_action) add command -label "Install" -underline 0 -accelerator "I" -state disabled -command "action_install"
    $w(popup_action) add separator
    $w(popup_action) add command -label "Edit info" -underline 0 -accelerator "E" -state normal -command "dialog_modal_package_info"

    bind $w(list_packages) 

    # scroll_packages
    set w(scroll_packages) [scrollbar $w(frame_packages).scroll -command "$w(list_packages) yview"]

    $w(list_packages) configure -yscrollcommand "$w(scroll_packages) set"

    # label_packages
    set w(label_packages) [label $w(frame_packages).label_packages -text "Packages" -font font_title -foreground "darkred"]

    # frame_actions
    set w(frame_actions) [frame $w(frame_packages).frame_actions -pady 10]

    set w(button_remove) [button $w(frame_actions).button_remove -text "Remove" -underline 0 -command "action_remove" -width 5]
    set w(button_install) [button $w(frame_actions).button_install -text "Install" -underline 0 -command "action_install" -width 5]
    set w(button_add) [button $w(frame_actions).button_add -text "Add..." -underline 0 -command "action_add" -width 5]

    # pane_vert add
    $w(pane_vert) add $w(frame_packages) -minsize 100
    $w(pane_vert) add $w(frame_console)

    # frame_status
    set w(frame_status) [frame $w(frame_main).frame_status -height 1 -background antiquewhite3]
    pack $w(frame_status) -fill x -anchor s -side bottom

    # packing: packages
    pack $w(label_packages) -side top -fill x
    pack $w(frame_actions) -side bottom -fill both
    pack $w(button_remove) -side left -padx 2
    pack $w(button_install) -side left -padx 2
    pack $w(button_add) -side left -padx 2
    pack $w(scroll_packages) -side right -fill y
    pack $w(list_packages) -fill both -expand 1

    # packing: console
    pack $w(label_console) -side top -fill x
    pack $w(scroll_console) -side right -fill y
    pack $w(console) -fill both -expand 1

    # packing: vertical pane container
    pack $w(pane_vert) -side top -fill both -expand 1

    # packing: main
    pack $w(frame_main) -fill both -expand 1
    
    # lock console
    console_lock
}

# the following dialog returns a filename or an empty string on failure

proc dialog_modal_add_browse { w_modal_add file_type } {
    global o
    global rundir

    set types {
	{ {source package} {.tar.gz .tar.bz2 .tar.Z .tgz .zip .shar.gz .shar.bz2 .shar.Z .shar.zip .shar .tar} }
	{ {gzip compressed tar archive} {.tar.gz .tgz} }
	{ {bzip2 compressed tar archive} {.tar.bz2} }
	{ {compress tar archive} {.tar.Z} }
	{ {zip archive} {.zip} }
	{ {gzip compressed shell archive} {.shar.gz} }
	{ {bzip2 compressed shell archive} {.shar.bz2} }
	{ {compress shell archive} {.shar.Z} }
	{ {zip compressed shell archive} {.shar.zip} }
	{ {shell archive} {.shar} }
	{ {tar archive} {.tar} }
	{ {all files} * }
    }

    set initial_dir ""
    if { $o(add_directory) != "" } {
	set initial_dir $o(add_directory)
    } else {
	set initial_dir $rundir
    }

    if { $file_type == "file" } {
        return [tk_getOpenFile -filetypes $types -parent $w_modal_add -title "Open package" -initialdir $initial_dir]

    } elseif { $file_type == "directory" } {
	return [tk_chooseDirectory -mustexist 1 -parent $w_modal_add -title "Open source directory" -initialdir $initial_dir]
    }

    return ""
}

proc is_valid_package { filename } {
    set retval 0
    set valid_ext [list .tar.gz .tar.bz2 .tar.Z .tgz .zip .shar.gz .shar.bz2 .shar.Z .shar.zip .shar .tar]

    if { [file_type $filename] == "directory" } {
	return 1
    }

    foreach thisone $valid_ext {
	if { [regexp "$thisone\$" $filename] } {
	    return 1
	}
    }

    return 0
}

proc unpack { filename } {
    set_status_bar_item "Unpack source" "yellow"
    set ext [file extension $filename]

    while { $ext == ".gz" || $ext == ".bz2" || \
		$ext == ".tgz" || $ext == ".Z" } {
	set rep ""
	console_print_comment "uncompressing $ext archive"
	
	switch $ext {
	    ".gz" {
		set cmd "/bin/gunzip -f"
	    }

	    ".tgz" {
		set cmd "/bin/gunzip -f"
		set rep ".tar"
	    }

	    ".Z" {
		set cmd "/bin/gunzip -f"
	    }

	    ".bz2" {
		set cmd "/usr/bin/bunzip2 -f"
	    }
	}
	
	if { [console_send_wait "$cmd $filename"] } {
	    return ""
	}

	set filename [regsub "$ext\$" $filename $rep]
	set ext [file extension $filename]
    }

    switch $ext {
	".tar" {
	    set cmd "/bin/tar -xvf $filename"
	}
	".shar" {
	    set cmd "/bin/sh $filename"
	}
	".zip" {
	    set cmd "/usr/bin/unzip -o $filename"
	}
    }

    set pattern {(\S+?)/}

    console_print_comment "unpacking $ext archive"

    if { [console_send_expect $cmd $pattern packdir] } {
	return ""
    }

    set_status_bar_item "Unpack source" "green"

    return $packdir
}

proc configure {} {
    global o
    global env
    global sigint_seen
    global package_info
    global builddir

    set title "Configuration"
    # generation scripts
    
    set_status_bar_item "Configure" "yellow"
    
    set configure_script ""
    set autoconf_version 0
    
    if { [file_type "configure"] == "file" && [file executable "configure"] } {
	# ok a nice configure script
	set configure_script "configure"
	
    } else {
	# maybe we have to bootstrap ?

	set autogen_scripts [list autogen.sh autogen bootstrap.sh bootstrap]
	
	foreach var $autogen_scripts {
	    set t [file_type $var]
	    if { $t == "file" } {
		# found something which looks good
		console_print_comment "configure generation script found"

		if { [console_send_wait "/bin/sh ./$var"] == 0 } {
		    if { [file_type "configure"] == "file" && [file executable "configure"] } {
			# ok a nice configure script was generated
			set configure_script "configure"
			break
		    }
		}
	    }
	}
    }

    if { $configure_script == "" } {
	# still no configure script found. We look in subdirectories.
	set configure_files [ find_files "." "-name configure" ]
	if { [llength $configure_files] > 0 } {
	    set message "A configure script for this package has not been found in the top source directory. However, configure scripts have been found in subdirectories.:\n\n"
	    foreach this $configure_files {
		append message "$this\n"
	    }
	    
	    append message "\nUse one of these configure scripts?\n"

	    set result [src_dialog $title $message question 0 \
			    "Yes" "No"]
	    if { $result == 0 } {
		set types {
		    { {configure script} {"configure"} }
		    { {all files} * }
		}
		set configure_script [tk_getOpenFile -filetypes $types -parent . -title "Choose configure script" -initialdir "$builddir/$package_info(name)"]
		if { $configure_script != "" } {
		    regsub "^$builddir/$package_info(name)/" $configure_script "" configure_script
		    set package_info(build_subdir) [file dirname $configure_script]
		    set configure_script [file tail $configure_script]
		    console_print_comment "change working directory to configuration directory"
		    console_print_command "cd $package_info(build_subdir)"
		    catch { cd $package_info(build_subdir) }

		}
	    }
	}
    }
    
    if { $configure_script != "" } {
	console_print_comment "configure software for this system"
	
	if { $o(manual_configure) && [set autoconf_version [get_autoconf_version $configure_script]] } {
	    set configure_flags [dialog_modal_configure $configure_script $autoconf_version]
	    
	} else {
	    # assume at least prefix is honoured
	    set configure_flags " --prefix=$o(prefix)"
	}
	
	set package_info(configured) "./$configure_script$configure_flags"

	if { $sigint_seen } {
	    src_warning $title "Configuration has been interrupted."
	    return 1
	}
	
	if { [console_send_wait "/bin/sh ./$configure_script$configure_flags"] } {
	    src_warning $title "Configuration NOT successful."
	    return 1
	}
    } else {
	# no configure script
	console_print_comment "configure script not available"
	set_status_bar_item "Configure" "red"
	set package_info(configured) "no"

	set result [src_dialog $title \
			"This package is not conforming: it does not provide a configure script, and no common configure-generation script is available. \nYour default installation prefix will be ignored, and you will not be able to configure the package at all. Really proceed?" \
			question \
			1 \
			"Yes" "No"]
	if { $result != 0 } {
	    return 1
	}
	return 0
    }
    
    set_status_bar_item "Configure" "green"
    
    return 0
}

proc make {} {
    # compile the code
    
    set title "Compilation"
    
    console_print_comment "compile software"
    set_status_bar_item "Compile" "yellow"
    
    set makefiles [list GNUmakefile makefile Makefile Makefile.unix Makefile.UNIX Makefile.Unix Makefile.linux Makefile.LINUX Makefile.Linux Makefile.cvs ]
    
    foreach var $makefiles {
	set t [file_type $var]
	if { $t == "file" } {
	    if { [console_send_wait "/usr/bin/make -f $var"] == 0 } {
		set_status_bar_item "Compile" "green"
		return $var
	    }
	}
    }
    
    src_warning $title "Could not compile the code."
    return ""
}

proc get_autoconf_version { configure } {
    set title "Configure script check"

    if { [catch { open $configure "r" } fh] } {
	src_warning $title $fh
	return 0
    }

    if { [catch { read $fh 512 } data] } {
	src_warning $title $data
	close $fh
	return 0
    }

    close $fh

    set lines [split $data "\n"]

    for { set i 1 } { $i < 5 } { incr i } {
	if { [regexp -nocase {generated.+autoconf.+?(\d+\.\d+).+} [lindex $lines $i] matchvar submatch ] } {
	    set version [expr double($submatch)]
	    console_print_comment "autoconf v. $version detected"
	    return $version
	}
    }
    return 0
}


# The following simple function tests a MAKEFILE (given as a file name) for
# a set of features. These features are passed in the form of a FEATURE_LIST,
# where each element is itself a list of two elements: the first is a
# descriptive FEATURE_NAME, completely unused but used for readability only
# while the second is a FEATURE_REGEX to match.
#
# The output is an ANSWER_LIST of boolean (0-1) values for each feature

proc test_makefile_features { makefile feature_list } {
    set title "Makefile feature test"
    set answer_list [list]

    if { [catch { open $makefile "r" } fh] } {
	src_warning $title $fh
	return 0
    }
    
    if { [catch { read $fh } data] } {
	src_warning $title $data
	close $fh
	return 0
    }
    
    close $fh

    set n [llength $feature_list]

    for { set i 0 } { $i < $n } { incr i } {
	set feature_name [lindex $feature_list $i 0]
	set feature_regex [lindex $feature_list $i 1]
	lappend answer_list [regexp -- $feature_regex $data]
    }

    return $answer_list
}

proc find_files { directory options } {
    global spawn_id

    set timeout -1
    set filelist [ list ]

    eval spawn -noecho "/bin/find $directory $options"
    update_actions

    expect {
	eof { 
	}

	-re ".+?\n" {
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    regsub -all "\n" $outtext "" outtext
	    regsub $directory $outtext "" outtext
	    
	    lappend filelist $outtext
	    exp_continue
	}
    }

    if { [wait_spawned] != 0 } {
	# FIXME: nothing for now
    }
	
    update_actions
    return $filelist
}

proc wait_spawned {} {
    global spawn_id
    global sigint_seen
    set res [wait]
    set spawn_id 0
    if { [ expr { [lindex $res 2] == 0 && [lindex $res 3] == 0 } ] } {
	if { $sigint_seen == 0 } {
	    return 0
	}
    }
    return 1
}

proc run_command_maybe_sudo { cmd user pass} {
    global w
    global sigint_seen
    upvar $user username
    upvar $pass password

    set permission_denied ""
    if { [console_send_expect $cmd "(Permission denied|Operation not permitted)" permission_denied] } {
	if { $permission_denied != "" } {

	    set userpass [dialog_modal_login]
	    if { $userpass == "" } {
		incr sigint_seen
		return 1
	    }
	    set username [lindex $userpass 0]
	    set password [lindex $userpass 1]
	    return [console_send_sudo $cmd $username $password]
	}
	return 1
    }
    return 0
}

proc make_install { makefile } {
    global destdir
    global package_info
    global o

    console_print_comment "install software"
    set_status_bar_item "Install" "yellow"

    set cmd "/usr/bin/make install"
    set feature_list {
	{ destdir_supported {\$\(DESTDIR\)} }
	{ strip_supported {install-strip *:} }
	{ descr_supported {description *:} }
	{ descr_long_supported {description-long *:} }
    }

    # the following is the foreach hack to work around the not existant
    # lassign command

    foreach {destdir_supported strip_supported descr_supported descr_long_supported} [test_makefile_features $makefile $feature_list] {}

    if { $descr_supported && $package_info(description) == "" } {
	expect_all "/usr/bin/make description" package_info(description)
    }

    if { $descr_long_supported && $package_info(long_description) == "" } {
	expect_all "/usr/bin/make description-long" package_info(long_description)
    }

    if { $strip_supported && $o(strip) == 1 } {
	append cmd "-strip"
    }

    if { $destdir_supported } {
	set cmd_destdir "$cmd DESTDIR=$destdir"
	console_print_comment "do a test install in a temporary directory"

	if { [console_send_wait $cmd_destdir] } {
	    console_print_comment "test install failed - DESTDIR violated?"
	    set destdir_supported 0

	} else {
	    set package_info(installed_files) [ find_files $destdir "" ]
	    console_print_comment "calc installation disk space usage in KB (approx)"
	    set pattern {(\d+)}

	    if { [console_send_expect "/usr/bin/du -ksx $destdir" $pattern package_info(installed_size)] } {
		set package_info(installed_size) ""
	    }
	    console_print_comment "now do the real install"
	}
    }

    if { $destdir_supported == 0 } {
	set result [src_dialog "Installation" \
			"This package is not conforming: package's Makefile does not honor DESTDIR.\nInformation about installed files will be missing, and future uninstall could be incomplete. Really proceed with installation?" \
			question \
			1 \
			"Yes" "No"]
	if { $result != 0 } {
	    return 1
	}

	if { $o(keep_source) == 0 } {
	    set result [src_dialog "Installation" \
			    "Installation without source code keeping has been requested.\nHowever, in order to be able to uninstall the non-conforming package, the only hope is using the configured source code to request uninstallation.\n\nDo you want to keep the source code as suggested?" \
			    question \
			    0 \
			    "Yes" "No"]
	    if { $result != 1 } {
		set o(keep_source) 1
	    }
	}
    }
    
    if { [run_command_maybe_sudo $cmd tmp1 tmp2] } {
	return 1
    }
    
    set_status_bar_item "Install" "green"
    
    return 0
}

proc keep_source { packdir } {
    global builddir
    global srcdir
    global o
    global package_info

    set title "Store source"

    set_status_bar_item "Store source" "yellow"
    console_print_comment "archive and compress configured source"

    console_print_command "cd $builddir"

    if { [catch { cd $builddir }] } {
	src_warning $title "Can not reach build directory."
	return ""
    }
    
    set srcname $package_info(name); append srcname ".tar"
    set cmd "/bin/tar -cf $srcname $packdir/"

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not archive source."
	return ""
    }

    switch $o(src_compress) {
	"bz2" {
	    set cmd "/usr/bin/bzip2 -9 -f $srcname"
	}

	"Z" {
	    set cmd "/usr/bin/compress -f $srcname"
	}

	"gz" {
	    set cmd "/bin/gzip -9 -f $srcname"
	}

	default {
	    set o(src_compress) "gz"
	    set cmd "/bin/gzip -9 -f $srcname"
	}
    }

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not compress source."
	return ""
    }

    append srcname ".$o(src_compress)"

    console_print_comment "move source package into source directory"
    set cmd "/bin/mv $srcname $srcdir/"

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not move package."
	return ""
    }

    set_status_bar_item "Store source" "green"
    return "$srcdir/$srcname"
}

proc add_package { filename packagename } {
    global package_info
    set title "Add package"

    array set package_info [new_package_info]
    set package_info(name) $packagename
    set basename [file tail $filename]

    if { ![file exists $filename] } {
	src_error $title "File not found: $filename ."
	return 1
    }

    if { ![is_valid_package $filename] } {
	src_error $title "Package format is not supported."
	return 1
    }

    console_clear
    console_print_title "Adding $basename\n\n"
    
    set retval [install_package $filename 1]
    clear_builddirs

    if { $retval == 0 } {
	registry_add_package [array get package_info]
	src_message $title "Package $package_info(name) successfully added."
    } else {
	src_warning $title "Package has NOT been added."
    }

    return $retval
}

proc install_package {fullname configureq} {
    global builddir
    global package_info
    global srcdir
    global o
    
    set title "Install package"

    set todo_list [list]
    set ft [file_type $fullname]

    if { $ft != "directory" } { set todo_list [list "Unpack source"] }
    if { $configureq } { lappend todo_list "Configure" }
    if { $o(install) } { lappend todo_list "Compile" "Install" }
    if { $o(keep_source) } { lappend todo_list "Store source" }

    show_status_bar $todo_list

    set basename [file tail $fullname]

    console_print_comment "change current directory to build directory"
    console_print_command "cd $builddir"

    if { [catch { cd $builddir }] } {
	src_warning $title "Can not reach build directory."
	hide_status_bar
	return 1
    }
    
    console_print_comment "copy source package into build directory"  
    set cmd "/bin/cp -R $fullname $builddir/$basename"

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not copy package into build directory."
	hide_status_bar
	return 1
    }

    if { $ft != "directory" } {
	set packdir [unpack $basename]
    } else {
	set packdir $basename
    }

    if { $packdir == "" } {
	src_warning $title "Unpack failed."
	hide_status_bar
	return 1
    }

    if { $package_info(name) == "" } {
	set package_info(name) $packdir
    }

    console_print_comment "change current directory to new package directory"
    console_print_command "cd $packdir"
    
    if { [catch { cd $packdir }] } {
	src_warning $title "Can not reach package directory."
	hide_status_bar
	return 1
    }

    if { $package_info(build_subdir) != "" } {
	console_print_comment "change current directory to build subdirectory"
	console_print_command "cd $package_info(build_subdir)"
	if { [catch { cd $package_info(build_subdir) }] } {
	    src_warning $title "Can not reach build subdirectory."
	    hide_status_bar
	    return 1
	}
    }

    if { $configureq } {

	if { [configure] } {
	    hide_status_bar
	    return 1
	}
    }

    if { $o(install) } {
	set makefile [make]

	if { $makefile == "" } {
	    hide_status_bar
	    return 1
	}

	if { [make_install $makefile] } {
	    hide_status_bar
	    return 1
	}

	set package_info(installed) "yes"
    }

    if { $o(keep_source) } {

	set package_info(source_location) [keep_source $packdir]
	
	if { $package_info(source_location) != "" } {
	    set package_info(source_available) "yes"
	    if { [catch { file size $package_info(source_location) } size] } {
		set package_info(source_size) ""
	    } else {
		set package_info(source_size) [expr $size/1024]
	    }
	}
    }
    
    hide_status_bar
    return 0
}

proc compare_dir_depth { dir1 dir2 } {
    set list1 [ split $dir1 ""]
    set val1 [llength [lsearch -all $list1 "/"]]
    set list2 [ split $dir2 ""]
    set val2 [llength [lsearch -all $list2 "/"]]

    return [expr $val1-$val2]
}

proc internal_uninstall {} {
    global package_info
    set user ""
    set pass ""

    if { $package_info(source_available) == "yes" } {
	set cross_answer 1
	console_print_comment "checking for files left over by make uninstall"
    } else {
	set cross_answer 0
    }

    set_status_bar_item "Internal uninstall" "yellow"
    
    # get a list of all file and link dependencies for the current package
    set dependencies [get_relative_dependencies]

    for { set i 0 } { $i < [llength $package_info(installed_files)] } { incr i } {
	
	set filename [lindex $package_info(installed_files) $i]
	set ft [file_ltype $filename]

	if { [expr { ($ft == "file" || $ft == "link") && [lsearch $dependencies $filename] == -1 }] } {

	    if { $cross_answer != 0 && $cross_answer != 3 } {
		set cross_answer [src_dialog "Make uninstall crosscheck" "A file left over by the make uninstall procedure has been detected:\n\n$filename\n\nIt does not seem to be claimed by any other installed package.\n\nShould it be deleted?\n" question $cross_answer "Always" "Yes" "No" "Never"]
	    }
	    
	    if { $cross_answer < 0 } {
		# closing the window is interpreted as a desperate action
		# so we never delete anything in this case
		set cross_answer 3
	    }
	    
	    if { $cross_answer < 2 } {
		console_print_comment "remove file left over by make uninstall"
		set permission_denied ""
		set cmd "/bin/rm -f $filename"
		if { $user == "" } {
		    if { [run_command_maybe_sudo $cmd user pass] } {
			return 1
		    }
		} else {
		    if { [console_send_sudo $cmd $user $pass] } {
			return 1
		    }
		}
		set package_info(installed_files) [lreplace $package_info(installed_files) $i $i]
		incr i -1
	    }
	}
    }
    	
    set_status_bar_item "Internal uninstall" "green"
    set_status_bar_item "Empty directories" "yellow"

    # now check empty directories

    set package_info(installed_files) [lsort -decreasing -command compare_dir_depth $package_info(installed_files)]
    set cross_answer 1
    
    for { set i 0 } { $i < [llength $package_info(installed_files)] } { incr i } {
	set filename [lindex $package_info(installed_files) $i]
	
	if { [file_ltype $filename] == "directory" } {
	    if { [ glob -nocomplain -directory $filename * ] == [list] } {
		if { $cross_answer != 0 && $cross_answer != 3 } {
		    set cross_answer [src_dialog "Make uninstall crosscheck" \
					  "Uninstallation left an empty directory:\n\n$filename\n\nShould it be deleted?\n" \
					  question \
					  $cross_answer \
					  "Always" "Yes" "No" "Never"]
		}
		
		if { $cross_answer < 0 } {
		    # even in this case, closing the dialog is interpreted
		    # as panic. We do not delete anything in this case
		    set cross_answer 3
		}
		
		if { $cross_answer < 2 } {
		    console_print_comment "remove empty directory"
		    set permission_denied ""
		    set cmd "/bin/rmdir $filename"
		    if { $user == "" } {
			if { [run_command_maybe_sudo $cmd user pass] } {
			    return 1
			}
		    } else {
			if { [console_send_sudo $cmd $user $pass] } {
			    return 1
			}
		    }
		    set package_info(installed_files) [lreplace $package_info(installed_files) $i $i]
		    incr i -1
		}
	    }
	}
    }
    
    set package_info(installed_files) [lsort -command compare_dir_depth $package_info(installed_files)]
    
    set_status_bar_item "Empty directories" "green"
    return 0
}

proc source_uninstall {} {
    global builddir
    global package_info
    set title "Remove source"

    set cmd "/bin/rm -f $package_info(source_location)"

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not remove source package."
	return 1
    }

    return 0
}

proc make_uninstall {} {
    global builddir
    global package_info

    set title "Uninstall"

    set fullname $package_info(source_location)
    set basename [file tail $fullname]

    console_print_comment "change current directory to build directory"
    console_print_command "cd $builddir"

    if { [catch { cd $builddir }] } {
	src_warning $title "Can not reach build directory."
	return 1
    }

    console_print_comment "copy source package into build directory"  
    set cmd "/bin/cp $fullname $builddir/$basename"

    if { [console_send_wait $cmd] } {
	src_warning $title "Could not copy package into build directory."
	return 1
    }

    set packdir [unpack $basename]

    if { $packdir == "" } {
	src_warning $title "Unpack failed."
	return 1
    }
    
    console_print_comment "change current directory to package directory"
    console_print_command "cd $packdir"
    if { [catch { cd $packdir }] } {
	src_warning $title "Can not reach package directory."
	return 1
    }

    if { $package_info(build_subdir) != "" } {
	console_print_comment "change current directory to build subdirectory"
	console_print_command "cd $package_info(build_subdir)"
	if { [catch { cd $package_info(build_subdir) }] } {
	    src_warning $title "Can not reach build subdirectory."
	    return 1
	}
    }
    
    set_status_bar_item "Uninstall" "yellow"

    console_print_comment "request uninstall"

    if { [run_command_maybe_sudo "/usr/bin/make uninstall" tmp1 tmp2] } {
	set_status_bar_item "Uninstall" "red"
	return 1
    }

    # remove from the list of installed files those that were successfully
    # removed by make uninstall

    for { set i 0 } { $i < [llength $package_info(installed_files)] } { incr i } {
	set filename [lindex $package_info(installed_files) $i]

	if { $filename == "" || [file_ltype $filename] == "none" } {
	    set package_info(installed_files) [lreplace $package_info(installed_files) $i $i]
	    incr i -1
	}
    }

    set_status_bar_item "Uninstall" "green"
    
    return 0
}

# the following is a dialog which returns a package name or the empty string
# writes global options: o(manual_configure), o(keep_source), o(install)

proc dialog_modal_add {} {
    global w
    global o
    global tips
    global dialog_retarr
    #   global icondir

    # radiobutton sets variables
    global add_choose_type
    global add_name_choice
    # initial radio choices
    set add_choose_type "file"
    set add_name_choice "auto"

    # these are the return values: the filename chosen and the package name
    # an empty value for filename means "STOP", an empty value for packagename
    # means "AUTO"
    array unset dialog_retarr
    array set dialog_retarr {
	filename ""
	packagename ""
    }

    set w(modal_add) [toplevel .modal_add]
    wm title $w(modal_add) "Choose source package"

    set w(add_frame_content) [frame $w(modal_add).frame_content]

    set w(add_frame_choose) [frame $w(add_frame_content).frame_choose -padx 10 -pady 10]
    set w(add_frame_type) [frame $w(add_frame_choose).frame_type]
    set w(add_radio_file) [radiobutton $w(add_frame_type).radio_file -text "Choose File:" -variable add_choose_type -value "file" -font font_small]
    tip $w(add_radio_file) $w(add_frame_choose) $tips(add_choose_file)

    set w(add_radio_dir) [radiobutton $w(add_frame_type).radio_dir -text "Choose Directory:" -variable add_choose_type -value "directory" -font font_small]
    tip $w(add_radio_dir) $w(add_frame_choose) $tips(add_choose_dir)

    set w(add_entry_choose) [entry $w(add_frame_choose).entry_choose -width 25]

    # -image image_folder_closed
    set w(add_button_choose) [button $w(add_frame_choose).button_choose -text "Browse" -width 4 -command { 
	set dialog_retarr(filename) [dialog_modal_add_browse $w(modal_add) $add_choose_type]
	$w(add_entry_choose) delete 0 end
	$w(add_entry_choose) insert 0 $dialog_retarr(filename)
	focus $w(add_button_ok)
    }]

    #    bind $w(add_button_choose) <Enter> {
    #	   %W configure -image image_folder_open
    #    }

    #    bind $w(add_button_choose) <Leave> {
    #      %W configure -image image_folder_closed
    #    }

    set w(add_frame_options) [labelframe $w(add_frame_content).frame_options -padx 10 -pady 10 -text "Options"]
    set w(add_check_configure) [checkbutton $w(add_frame_options).check_configure -text "manual configuration" -variable o(manual_configure)]
    tip $w(add_check_configure) $w(add_frame_content) $tips(manual_configure)

    set w(add_check_install) [checkbutton $w(add_frame_options).check_install -text "install" -variable o(install)]
    tip $w(add_check_install) $w(add_frame_content) $tips(install)

    set w(add_check_source) [checkbutton $w(add_frame_options).check_source -text "keep source" -variable o(keep_source)]
    tip $w(add_check_source) $w(add_frame_content) $tips(keep_source)

    set w(add_frame_name) [labelframe $w(add_frame_content).frame_name -padx 10 -pady 10 -text "Package name"]
    set w(add_frame_name2) [frame $w(add_frame_name).frame_name2]

    set w(add_radio_name_auto) [radiobutton $w(add_frame_name2).radio_name_auto -text "Assign default name" -variable add_name_choice -value "auto" -font font_text -command {dialog_modal_add_name_check $add_name_choice}]
    tip $w(add_radio_name_auto) $w(add_frame_content) $tips(package_name_auto)

    set w(add_radio_name_manual) [radiobutton $w(add_frame_name2).radio_name_manual -text "Custom name" -variable add_name_choice -value "manual" -font font_text -command {dialog_modal_add_name_check $add_name_choice}]
    tip $w(add_radio_name_manual) $w(add_frame_content) $tips(package_name_manual)

    set w(add_entry_name) [entry $w(add_frame_name2).entry_name -width 16 -state readonly -validate key -vcmd { is_valid_filestring %P }]

    tip $w(add_entry_name) $w(add_frame_content) $tips(package_name)

    set w(add_frame_actions) [frame $w(modal_add).frame_actions -padx 10 -pady 10]
    set w(add_button_ok) [button $w(add_frame_actions).button_ok -text "Ok" -width 5 -command { 
	set dialog_retarr(filename) [file normalize [$w(add_entry_choose) get]]
	if { $add_name_choice == "manual" } {
	    set dialog_retarr(packagename) [$w(add_entry_name) get]
	}
	destroy $w(modal_add)
    }]

    set w(add_button_cancel) [button $w(add_frame_actions).button_cancel -text "Cancel" -width 5 -command { 
	set dialog_retarr(filename) ""; destroy $w(modal_add)
    }]
    
    pack $w(add_frame_content) -side top -fill x
    pack $w(add_frame_choose) -side top -fill x -expand 1
    pack $w(add_frame_type) -side left
    pack $w(add_radio_file) -side top -anchor w
    pack $w(add_radio_dir) -side top -anchor w
    pack $w(add_entry_choose) -side left
    pack $w(add_button_choose) -side left -padx 4

    pack $w(add_frame_options) -side left -fill x -expand 1
    pack $w(add_frame_name) -side left -fill both -expand 1

    pack $w(add_check_configure) -side top -anchor w
    pack $w(add_check_install) -side top -anchor w
    pack $w(add_check_source) -side top -anchor w

    pack $w(add_frame_name2) -side left -fill x
    pack $w(add_radio_name_auto) -side top -anchor w
    pack $w(add_radio_name_manual) -side top -anchor w
    pack $w(add_entry_name) -side right

    pack $w(add_frame_actions) -side top
    pack $w(add_button_ok) -side left -padx 8
    pack $w(add_button_cancel) -side left -padx 8

    focus $w(add_button_choose)
    dialog_wait $w(modal_add) 0
    
    # pseudocode: return $dialog_retarr
    return 0
}

# handle activation of text entry on selection of radio buttons

proc dialog_modal_add_name_check { add_name_choice  } {
    global w

    if { $add_name_choice == "auto" } {
	$w(add_entry_name) configure -state readonly

    } else {
	$w(add_entry_name) configure -state normal
    }
}

proc lock_actions {} {
    global actions_locked
    global sigint_seen

    set actions_locked 1
    set sigint_seen 0
}

proc unlock_actions {} {
    global actions_locked
    global package_info
    global w

    set actions_locked 0
    array set package_info [ new_package_info ]

    $w(list_packages) configure -state normal
    $w(list_packages) selection clear 0 end
}


# console must be unlocked prior to any operation on the console widget.
# At the end of the operations, it must be locked again.
# This is the only way I could simulate a readonly text console.
# text widget should be fixed. If you know better, please tell me.

proc console_unlock {} {
    global w
    $w(console) configure -state normal
}

proc console_lock {} {
    global w
    $w(console) configure -state disabled
}

proc console_home {} {
    global w
    console_unlock
    $w(console) see 0.0
    console_lock
}

proc console_clear {} {
    global w

    console_unlock
    set tag_list [$w(console) tag names]

    if { [llength $tag_list] > 0 } {
	$w(console) tag delete $tag_list
    }

    $w(console) delete 0.0 end
    console_lock
}

proc console_print_title {some_text} {
    global w

    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end "$some_text"
    $w(console) tag add tag_title mark_tmp insert
    $w(console) mark unset mark_tmp
    console_lock
}

proc console_print_subtitle {some_text} {
    global w

    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end $some_text
    $w(console) tag add tag_subtitle mark_tmp insert
    $w(console) mark unset mark_tmp
    console_lock
}

proc console_print_ref {refname some_text script} {
    global w

    $w(console) tag configure "tag_ref_$refname" -font font_subtitle 

    $w(console) tag bind "tag_ref_$refname" <ButtonRelease> $script

    $w(console) tag bind "tag_ref_$refname" <Enter> "
	$w(console) configure -cursor left_ptr
	$w(console) tag configure tag_ref_$refname -background \#ffffc0
    "

    $w(console) tag bind "tag_ref_$refname" <Leave> "
	$w(console) configure -cursor xterm
	$w(console) tag configure tag_ref_$refname -background antiquewhite
    "
    
    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end $some_text
    $w(console) tag add "tag_ref_$refname" mark_tmp insert
    $w(console) mark unset mark_tmp
    console_lock
}


proc console_print {some_text} {
    global w

    console_unlock
    $w(console) insert end $some_text
    $w(console) yview -pickplace end
    console_lock
}

proc console_print_code {some_text} {
    global w

    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end $some_text
    $w(console) tag add tag_code mark_tmp insert
    $w(console) mark unset mark_tmp

    $w(console) yview -pickplace end
    console_lock
}

proc console_print_command {some_text} {
    global w
    
    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end "\$ $some_text\n"
    $w(console) tag add tag_command mark_tmp insert
    $w(console) mark unset mark_tmp
    $w(console) yview -pickplace end
    console_lock

    status_print_command $some_text
}

proc console_print_err {some_text} {
    global w
    
    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end "$some_text"
    $w(console) tag add tag_err mark_tmp insert
    $w(console) mark unset mark_tmp
    $w(console) yview -pickplace end
    console_lock
}

proc console_print_comment {some_text} {
    global w
    
    console_unlock
    $w(console) mark set mark_tmp insert
    $w(console) mark gravity mark_tmp left
    $w(console) insert end "\# $some_text\n"
    $w(console) tag add tag_comment mark_tmp insert
    $w(console) mark unset mark_tmp
    $w(console) yview -pickplace end
    console_lock
}

proc console_intro {} {
    global packages

    console_clear
    console_print_title "Welcome to Source Installer!\n\n"
    console_print "\tSource Installer automates configuration, compilation, installation and removal of source packages (.tar.gz, .tar.bz2, ...), using the common procedures used to install Unix software.\n\n\tTo install a new source package, choose 'Add'.\nAll commands, output, warnings and errors will be shown in this console.\n\n"
    console_home
}

proc console_send_wait {string} {
    global spawn_id

    # global timeout
    # puts "timeout: $timeout"
    # exit 1

    set timeout -1

    console_print_command "$string"

    if { [catch { eval spawn -noecho $string }] } {
	return 0
    }

    update_actions

    expect {
#	timeout {
#	    puts "timeout."
#	    exit 1
#	}

	eof {
	}

	-re ".+" {
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    console_print $outtext
	    exp_continue
	}
    }

    set res [wait_spawned]
    update_actions

    return $res
}

proc console_send_sudo {string username password} {
    global spawn_id
    set timeout -1

    set string "\"$string\""

    console_print_command "/bin/su $username -c $string"

    eval spawn -noecho /bin/su $username -c $string
    update_actions

    set password_sent 0

    expect {
	eof {
	}

	"assword:" {
	    if { $password_sent == 0 } {
		sleep 1
		exp_send "$password\r"
		set password_sent 1
	    }
	    exp_continue
	}

	-re ".+" {
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    console_print $outtext
	    exp_continue
	}
    }

    set res [wait_spawned]
    update_actions

    return $res
}

proc console_send_expect {cmd pattern f} {
    global spawn_id
    upvar $f fetch
    set timeout -1

    set fetch ""

    console_print_command "$cmd"
    if { [catch { eval spawn -noecho $cmd }] } {
	return 0
    }

    update_actions

    expect {

	eof {
	}

	-re $pattern {
	    if { $fetch == "" } {
		set fetch $expect_out(1,string)
	    }
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    console_print $outtext
	    exp_continue
	}

	-re ".+" {
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    console_print $outtext
	    exp_continue
	}
    }

    set res [wait_spawned]
    update_actions

    return $res
}

proc expect_all {cmd f} {
    global spawn_id
    upvar $f fetch
    set timeout -1

    set fetch ""

    eval spawn -noecho $cmd
    update_actions

    expect {
	eof { 
	}

	-re ".+" {
	    regsub -all "\r" $expect_out(buffer) "" outtext
	    append fetch $outtext
	    exp_continue
	}
    }

    set res [wait_spawned]
    update_actions

    return $res
}

proc dialog_modal_configure_check { feature } {
    global dialog_retarr
    global w

    if { $dialog_retarr($feature) == 1 } {
	$w("entry_$feature") configure -state normal
	focus $w("entry_$feature")
    } else {
	$w("entry_$feature") configure -state normal
	$w("entry_$feature") delete 0 end
	$w("entry_$feature") configure -state readonly
    }
}

# this dialog reads an autoconf generated configure script to determine supported
# configure flags and allows the user to manually change them
# the return value is a string with the configure flags

proc dialog_modal_configure { configure_script autoconf_version } {
    global dialog_retarr
    global dialog_retval
    global w
    global o
    global sigint_seen
    global env

    # this is the return value - by default interrupt operation
    set dialog_retval ""
    array unset dialog_retarr
    array set dialog_retarr {}

    set cmd "/bin/sh ./$configure_script --help"
    set output ""

    if { [expect_all $cmd output] } {
	# seems that configure script is not so gnu after all, proceeding
	# without flags
	return ""
    }
    
    set output [split $output "\n"]

    # the following could be easily read from the script too
    # but I would like to extend this in the future to detect non-autoconf
    # generated configure scripts that DO support any of these

    array set base {
	"--prefix" { "" "where to install files [default=/usr/local]" }
	"--exec-prefix"	{ "" "where to install arch dependant files [default=PREFIX]" }
	"--bindir" { "" "user executables [EPREFIX/bin]" }
	"--sbindir" { "" "system admin executables [EPREFIX/sbin]" }
	"--libexecdir" { "" "program executables [EPREFIX/libexec]" }
	"--datadir" { "" "read-only architecture-independent data [PREFIX/share]" }
	"--sysconfdir" { "" "read-only single-machine data [PREFIX/etc]" }
	"--sharedstatedir" { "" "modifiable architecture-independent data [PREFIX/com]" }
	"--localstatedir" { "" "modifiable single-machine data [PREFIX/var]" }
	"--libdir" { "" "object code libraries [EPREFIX/lib]" }
	"--includedir" { "" "C header files [PREFIX/include]" }
	"--oldincludedir" { "" "C header files for non-gcc [/usr/include]" }
	"--infodir" { "" "info documentation [PREFIX/info]" }
	"--mandir" { "" "man documentation [PREFIX/man]" }
    }

    if { $autoconf_version >= 2.50 } {
	array set keywords {
	    optional_features {
		"^Optional Features" {^  (--[^= \[]+)(\[?=\w+\]?)?[ \t]*(.*)}
	    }
	    optional_packages {
		"^(Optional Packages|X features)" 
		{^  (--[^= \[]+)(\[?=\w+\]?)?[ \t]*(.*)}
	    }
	    environment_variables {
	    "^Some influential environment variables"
		{^  (\w+)(\[?=\w+\]?)?[ \t]*(.*)}
	    }
	}
	array set optional_features {}
	array set optional_packages {}
	array set environment_variables {}
	
    } else {
	array set keywords {
	    optional_features {
		"^--enable and --with options recognized"
		{^  (--[^= \[]+)(\[?=\w+\]?)?[ \t]*(.*)}
	    }
	}
	array set optional_features {}
    }
	
    set feature_sets [list base optional_features optional_packages environment_variables ]

    set lineno 0
    set n [llength $output]

    for { set lineno 0 } { $lineno < $n } { incr lineno } {
	foreach keyword [array names keywords] {
	    set line [lindex $output $lineno]
	    
	    if { [regexp -nocase [lindex $keywords($keyword) 0] $line] } {
		set pattern [lindex $keywords($keyword) 1]
		#		puts "matched regexp: $keywords($keyword)"
		while { 1 } {
		    incr lineno
		    if { $lineno >= $n } { break }

		    set line [lindex $output $lineno]
		    if { ![regexp $pattern $line matchvar submatch paramet comment] } {
			break
		    }
		    
		    if { [regexp ".+-PACKAGE" $submatch] } {
			continue
		    }
		    if { [regexp ".+-FEATURE" $submatch] } {
			continue
		    }

		    set fwidx 0

		    while { 1 } {
			incr fwidx
			if { [expr $lineno+$fwidx >= $n ] } { break }
			set fwline [lindex $output [expr $lineno+$fwidx]]
			if { $fwline == "" } { break }
			if { [regexp $pattern $fwline] } { break }

			if { ![regexp {^[ \t]+(.+)} $fwline matchvar submatch2] } {
			    break
			}
			append comment " $submatch2"
		    }

		    set "$keyword\($submatch\)" [ list "" $comment ]
		    incr lineno [expr $fwidx-1]
		}
		if { $lineno >= $n } { break }
	    }
	}
    }

    # sorry about the ugliness above, if you can think of a better way
    # please help

    # so now we have extended_features enriched with options taken
    # from configure help

    # now build dialog leaving out the comments for the moment

    set w(modal_conf_elder) [toplevel .modal_conf_elder]
    wm title $w(modal_conf_elder) "Configure package"

    set w(modal_conf) [ frame $w(modal_conf_elder).modal_conf ]

    pack $w(modal_conf)

    set w(conf_frame_actions) [frame $w(modal_conf_elder).frame_actions -padx 10 -pady 10]
    pack $w(conf_frame_actions) -side bottom

    set w(conf_frame_base) [labelframe $w(modal_conf).frame_base -padx 10 -pady 10 -text "Install Directories"]
    pack $w(conf_frame_base) -side left -fill both

    set w(conf_frame_optional_features) [labelframe $w(modal_conf).frame_optional_features -padx 10 -pady 10 -text "Optional Features"]
    pack $w(conf_frame_optional_features) -side left -fill both

    set w(conf_frame_optional_packages) [labelframe $w(modal_conf).frame_optional_packages -padx 10 -pady 10 -text "Optional Packages"]
    pack $w(conf_frame_optional_packages) -side left -fill both

    set w(conf_frame_x_features) [labelframe $w(modal_conf).frame_x_features -padx 10 -pady 10 -text "X Features"]
    pack $w(conf_frame_x_features) -side left -fill both

    set w(conf_frame_environment_variables) [labelframe $w(modal_conf).frame_environment_variables -padx 10 -pady 10 -text "Environment Variables"]
    pack $w(conf_frame_environment_variables) -side left -fill both

    # note (again) x_features have been redirected to optional_packages

    foreach featureset $feature_sets {
	set feature_list [lsort [array names $featureset]]
	set featcount [llength $feature_list]

	if { $featcount > 20 } {
	    set five [set w(scroll_${featureset}) [scrollbar $w(conf_frame_${featureset}).scroll_${featureset} -command [ list dialog_modal_configure_scroll $featureset $feature_list]]]
	    pack $five -side left -fill y
	    $w(scroll_${featureset}) set 0.0 [expr 19.0/$featcount]
	}

	set i 0
	foreach feature $feature_list {
	    set zero [set w("frame_$feature") [frame "$w(conf_frame_$featureset).frame_$i" -padx 5 -pady 1]]

	    # one and two reserved for comment or future stuff
	    
	    set three [set w("entry_$feature") [entry "$zero.entry_$i" -width 10 -state readonly -font font_text]]

	    set four [set w("check_$feature") [checkbutton "$zero.check_$i" -variable dialog_retarr($feature) -font font_text -text "$feature" -justify left -command [ list dialog_modal_configure_check $feature ]]]

	    set com [eval "lindex \$$featureset\($feature\) 1"]
	    tip $four "$w(conf_frame_$featureset)" $com
	    
	    if { $feature == "--prefix" } {
		$four configure -relief groove -background \#ffffa0
		set dialog_retarr($feature) 1
		$three configure -state normal
		$three insert 0 $o(prefix)

	    } elseif { $featureset == "environment_variables" } {
		if [info exists env($feature)] {
		    $three configure -state normal
		    $three insert 0 $env($feature)
		    $three configure -state readonly
		}
	    }

	    pack $zero -side top -fill x
	    pack $four -side left -fill x
	    pack $three -side right

	    incr i
	}
	dialog_modal_configure_showfeats $featureset $feature_list 0
    }

    set w(conf_button_ok) [button $w(conf_frame_actions).button_ok -text "Ok" -width 5 -command {
	set dialog_retval ""
	foreach var [array names dialog_retarr] {
	    if { $dialog_retarr($var) != 0 } {
		set tmp [$w("entry_$var") get]

		if { $tmp == "" && ![ regexp {^--} $var ] } {
		    continue
		}

		append dialog_retval " \"$var"
		if { $tmp != "" } {
		    append dialog_retval "=$tmp"
		}
		append dialog_retval "\""
	    }
	}
	
	destroy $w(modal_conf_elder)
    }]

    set w(conf_button_auto) [button $w(conf_frame_actions).button_auto -text "Auto" -width 5 -command {
	set dialog_retval " --prefix=$o(prefix)"
	destroy $w(modal_conf_elder)
    }]

    set w(conf_button_stop) [button $w(conf_frame_actions).button_stop -text "Stop" -width 5 -command {
	set dialog_retval ""
	incr sigint_seen
	destroy $w(modal_conf_elder)
    }]

    pack $w(conf_button_ok) -side left -padx 8 -pady 8
    pack $w(conf_button_auto) -side left -padx 8 -pady 8
    pack $w(conf_button_stop) -side left -padx 8 -pady 8

    focus $w(conf_button_ok)
    dialog_wait $w(modal_conf_elder) 1
    
    return $dialog_retval
}

proc dialog_modal_configure_showfeats { featureset feature_list start } {
    global w

    foreach feature $feature_list {
	if { $feature == "" } { break }
	pack forget $w("frame_$feature")
    }

    set featcount [llength $feature_list]

    for { set i $start } { $i < [expr $start+20] } { incr i } {
	if { $i >= $featcount } { break }
	set feature [lindex $feature_list $i]
	if { $feature == "" } { break }

	pack $w("frame_$feature") -side top -fill x
	pack $w("check_$feature") -side left -fill x
	pack $w("entry_$feature") -side right
    }
}

proc dialog_modal_configure_scroll args {
    global w

    set featureset [lindex $args 0]
    set feature_list [lindex $args 1]
    set cmd [lindex $args 2]

    set featcount [llength $feature_list]

    if { $cmd == "moveto" } {
	set idx [expr [lindex $args 3] * $featcount]
	set idx [expr int($idx)]

	if { $idx < 0 } {
	    set idx 0
	} elseif { [expr $idx+20] > $featcount } {
	    set idx [expr $featcount-20]
	}
	
    } else {
	# scroll
	set step [lindex $args 3]

	if { [lindex $args 4] == "pages" } {
	    set step [expr 20 * $step]
	}

	set first [lindex [$w(scroll_${featureset}) get] 0]
	set idx [expr int($first * $featcount)]
	incr idx $step

	if { $idx < 0 } {
	    set idx 0
	} elseif { [expr $idx+20] > $featcount } {
	    set idx [expr $featcount-20]
	}
    }

    set first [expr $idx / double([expr $featcount-1])]
    set last [expr [expr $idx+19] / double([expr $featcount-1])]

    $w(scroll_${featureset}) set $first $last
    dialog_modal_configure_showfeats $featureset $feature_list $idx
} 


# this dialog lets the user decide which parts of the package to remove
# (source or final installation or both)
# return value is 0=remove nothing, 1=uninstall, 2=remove source, 3=remove all

proc dialog_modal_remove {} {
    global w
    global o
    global dialog_retval
    global dialog_retarr
    global package_info

    # this is the return value
    set dialog_retval 0
    array unset dialog_retarr
    array set dialog_retarr {
	uninstall 0
	unsrc 0
    }

    set w(modal_remove) [toplevel .modal_remove]
    wm title $w(modal_remove) "Remove package"

    set w(remove_frame_options) [frame $w(modal_remove).frame_options -padx 10 -pady 10]
    set w(remove_label_options) [label $w(remove_frame_options).label_options -text "Remove package $package_info(name)?"]
    
    if { $package_info(installed) == "yes" } {
	set w(remove_check_uninstall) [checkbutton $w(remove_frame_options).check_uninstall -text "uninstall package (remove installed files)" -variable dialog_retarr(uninstall)]
	set dialog_retarr(uninstall) 1
    }

    if { $package_info(source_available) == "yes" } {
	set w(remove_check_unsrc) [checkbutton $w(remove_frame_options).check_unsrc -text "remove source code" -variable dialog_retarr(unsrc)]
	set dialog_retarr(unsrc) 1
    }

    set w(remove_frame_actions) [frame $w(modal_remove).frame_actions -padx 10 -pady 10]
    set w(remove_button_ok) [button $w(remove_frame_actions).button_ok -text "Ok" -width 5 -command { 
	set dialog_retval 0
	if { $dialog_retarr(uninstall) } { incr dialog_retval 1 }
	if { $dialog_retarr(unsrc) } { incr dialog_retval 2 }
	destroy $w(modal_remove)
    }]

    set w(remove_button_cancel) [button $w(remove_frame_actions).button_cancel -text "Cancel" -width 5 -command { 
	set dialog_retval 0
	destroy $w(modal_remove)
    }]
    
    pack $w(remove_frame_options) -side top -fill x
    pack $w(remove_label_options) -side top -anchor nw

    if { $package_info(installed) == "yes" } {
	pack $w(remove_check_uninstall) -side top -anchor w -padx 8 -pady 4
    }

    if { $package_info(source_available) == "yes" } {
	pack $w(remove_check_unsrc) -side top -anchor w -padx 8 -pady 4
    }

    pack $w(remove_frame_actions) -side top
    pack $w(remove_button_ok) -side left -padx 8
    pack $w(remove_button_cancel) -side left -padx 8

    focus $w(remove_button_ok)
    dialog_wait $w(modal_remove) 0
    
    return $dialog_retval
}

# this is the preferences dialog
# from here the user changes and saves preferences

proc dialog_modal_preferences { } {
    global dialog_retval
    global dialog_retarr
    global o
    global tips
    global w
    global src_compress_formats

    set dialog_retval 0
    array unset dialog_retarr
    array set dialog_retarr {}

    set w(modal_pref) [toplevel .modal_pref]
    wm title $w(modal_pref) "Preferences"

    set w(pref_frame_options) [labelframe $w(modal_pref).frame_options -padx 10 -pady 10 -text "Default installation choices"]

    set dialog_retarr(manual_configure) $o(manual_configure)
    set dialog_retarr(install) $o(install)
    set dialog_retarr(keep_source) $o(keep_source)
    set dialog_retarr(strip) $o(strip)

    set w(pref_check_configure) [checkbutton $w(pref_frame_options).check_configure -text "manual configuration" -variable dialog_retarr(manual_configure)]
    tip $w(pref_check_configure) $w(pref_frame_options) $tips(manual_configure)

    set w(pref_check_install) [checkbutton $w(pref_frame_options).check_install -text "install" -variable dialog_retarr(install)]
    tip $w(pref_check_install) $w(pref_frame_options) $tips(install)

    set w(pref_check_source) [checkbutton $w(pref_frame_options).check_source -text "keep source" -variable dialog_retarr(keep_source)]
    tip $w(pref_check_source) $w(pref_frame_options) $tips(keep_source)
    
    set w(pref_check_strip) [checkbutton $w(modal_pref).check_strip -text "strip binaries (not recommended)" -variable dialog_retarr(strip)]
    tip $w(pref_check_strip) $w(pref_frame_options) $tips(strip)
    
    set w(pref_frame_align) [ frame $w(modal_pref).frame_align ]

    set w(pref_frame_dirs) [labelframe $w(pref_frame_align).frame_dirs -text "Default directories" -padx 10 -pady 10]

    set w(pref_frame_prefix) [frame $w(pref_frame_dirs).frame_prefix]
    set w(pref_label_prefix) [label $w(pref_frame_prefix).label_prefix -text "prefix"]
    tip $w(pref_label_prefix) $w(pref_frame_options) $tips(prefix)
    set w(pref_entry_prefix) [entry $w(pref_frame_prefix).entry_prefix -width 15]
    $w(pref_entry_prefix) insert 0 $o(prefix)

    set w(pref_frame_adddir) [frame $w(pref_frame_dirs).frame_adddir]
    set w(pref_label_adddir) [label $w(pref_frame_adddir).label_adddir -text "add directory"]
    tip $w(pref_label_adddir) $w(pref_frame_options) $tips(add_directory)
    set w(pref_entry_adddir) [entry $w(pref_frame_adddir).entry_adddir -width 15]
    $w(pref_entry_adddir) insert 0 $o(add_directory)

    set w(pref_frame_compress) [labelframe $w(pref_frame_align).frame_compress -padx 10 -pady 10 -text "Src compression"]
    tip $w(pref_frame_compress) $w(pref_frame_options) $tips(src_compress)

    set w(pref_list_compress) [listbox $w(pref_frame_compress).list_compress -listvariable src_compress_formats -exportselection 0 -width 6 -height [llength $src_compress_formats]]

    for { set i 0 } { $i < [llength $src_compress_formats] } { incr i } {
	if { $o(src_compress) == [lindex $src_compress_formats $i] } {
	    $w(pref_list_compress) selection set $i
	    break
	}
    }

    set w(pref_frame_actions) [frame $w(modal_pref).frame_actions -padx 10 -pady 10]
    
    set w(pref_button_save) [button $w(pref_frame_actions).button_save -text "Save" -width 5 -command { 
	set o(manual_configure) $dialog_retarr(manual_configure)
	set o(install) $dialog_retarr(install)
	set o(keep_source) $dialog_retarr(keep_source)
	set o(strip) $dialog_retarr(strip)
	set o(prefix) [$w(pref_entry_prefix) get]
	set o(add_directory) [$w(pref_entry_adddir) get]
	set o(src_compress) [lindex $src_compress_formats [$w(pref_list_compress) curselection]]
	options_save
	destroy $w(modal_pref)
    }]

    set w(pref_button_reset) [button $w(pref_frame_actions).button_reset -text "Defaults" -command {
	array set o [new_options]
	options_save
	destroy $w(modal_pref)
    }]
    
    set w(pref_button_cancel) [button $w(pref_frame_actions).button_cancel -text "Cancel" -width 5 -command {
	destroy $w(modal_pref)
    }]

    pack $w(pref_frame_options) -side top -anchor w -fill x -pady 10
    pack $w(pref_check_configure) -side top -anchor w
    pack $w(pref_check_install) -side top -anchor w
    pack $w(pref_check_source) -side top -anchor w
    pack $w(pref_check_strip) -side top -anchor w -padx 12

    pack $w(pref_frame_align) -side top

    pack $w(pref_frame_dirs) -side left -anchor w -pady 10 -fill y

    pack $w(pref_frame_prefix) -fill x
    pack $w(pref_label_prefix) -side left -padx 4 -anchor w
    pack $w(pref_entry_prefix) -side right

    pack $w(pref_frame_adddir) -fill x
    pack $w(pref_label_adddir) -side left -padx 4 -anchor w
    pack $w(pref_entry_adddir) -side right

    pack $w(pref_frame_compress) -side left -anchor w -pady 10 -fill y
    pack $w(pref_list_compress) -side left -padx 4
    
    pack $w(pref_frame_actions) -side top
    pack $w(pref_button_save) -side left -padx 8
    pack $w(pref_button_reset) -side left -padx 8
    pack $w(pref_button_cancel) -side left -padx 8

    focus $w(pref_button_cancel)
    dialog_wait $w(modal_pref) 0
    
    return $dialog_retval
}

# this is the package information edit dialog
# from here the user changes package name, short and long descriptions

proc dialog_modal_package_info { } {
    global dialog_retval
    global dialog_retarr
    global tips
    global w
    global package_info
    global srcdir

    set dialog_retval 0
    array unset dialog_retarr
    array set dialog_retarr {}

    set w(modal_packinfo) [toplevel .modal_packinfo]
    set title "Edit package information"
    wm title $w(modal_packinfo) $title

    set w(packinfo_frame_fields) [frame $w(modal_packinfo).frame_fields -padx 10 -pady 10]

    set w(packinfo_frame_name) [labelframe $w(packinfo_frame_fields).frame_name -padx 10 -pady 10 -text "package name"]
    set w(packinfo_entry_name) [entry $w(packinfo_frame_name).entry_name -width 50]
    $w(packinfo_entry_name) insert 0 $package_info(name)
    $w(packinfo_entry_name) configure -validate key -vcmd { is_valid_filestring %P }
    tip $w(packinfo_entry_name) $w(packinfo_frame_name) $tips(package_name)

    set w(packinfo_frame_descr) [labelframe $w(packinfo_frame_fields).frame_descr -padx 10 -pady 10 -text "package short description"]
    set w(packinfo_entry_descr) [entry $w(packinfo_frame_descr).entry_descr -width 50]
    $w(packinfo_entry_descr) insert 0 $package_info(description)
    tip $w(packinfo_entry_descr) $w(packinfo_frame_descr) $tips(package_description)

    set w(packinfo_frame_descr_long) [labelframe $w(packinfo_frame_fields).frame_descr_long -padx 10 -pady 10 -text "package long description"]
    set w(packinfo_text_descr_long) [text $w(packinfo_frame_descr_long).text_descr_long -width 50 -height 10]
    $w(packinfo_text_descr_long) insert end $package_info(long_description)
    tip $w(packinfo_text_descr_long) $w(packinfo_frame_descr_long) $tips(package_long_description)

    set w(packinfo_frame_actions) [frame $w(modal_packinfo).frame_actions -padx 10 -pady 10]
    
    set w(packinfo_button_save) [button $w(packinfo_frame_actions).button_save -text "Save" -width 5 -command { 

	array set new_package_info [array get package_info]
	set new_package_info(name) [$w(packinfo_entry_name) get]
	set new_package_info(description) [$w(packinfo_entry_descr) get]
	set new_package_info(long_description) [$w(packinfo_text_descr_long) get 0.0 end]
	if { $new_package_info(name) == "" } {
	    src_warning $title "The package name can not be empty."
	} else {
	    if { $package_info(name) != $new_package_info(name) } {
		if { $package_info(source_location) != "" } {
		    set old_src_name [file tail $package_info(source_location)]
		    set new_src_name ""
		    regsub "$package_info(name)" $old_src_name $new_package_info(name) new_src_name
		    set new_package_info(source_location) "$srcdir/$new_src_name"
		}
	    }

	    if { [registry_add_package [array get new_package_info]] == 0 } {
		if { $package_info(name) != $new_package_info(name) } {
		    if { $package_info(source_location) != "" } {
			set cmd "/bin/mv $package_info(source_location) $new_package_info(source_location)"
			if { [console_send_wait $cmd] } {
			    src_warning $title "Could not rename package source!"
			    set new_package_info(source_location) $package_info(source_location)
			}
		    }
		    registry_remove_package [array get package_info]
		}
		array set package_info [array get new_package_info]
	    }
	    
	    destroy $w(modal_packinfo)
	}
    }]

    set w(packinfo_button_cancel) [button $w(packinfo_frame_actions).button_cancel -text "Cancel" -width 5 -command {
	destroy $w(modal_packinfo)
    }]

    pack $w(packinfo_frame_fields) -side top -anchor w -fill x -pady 10

    pack $w(packinfo_frame_name) -fill x
    pack $w(packinfo_entry_name)

    pack $w(packinfo_frame_descr) -fill x
    pack $w(packinfo_entry_descr)

    pack $w(packinfo_frame_descr_long) -fill x
    pack $w(packinfo_text_descr_long)
    
    pack $w(packinfo_frame_actions) -side top
    pack $w(packinfo_button_save) -side left -padx 8
    pack $w(packinfo_button_cancel) -side left -padx 8

    focus $w(packinfo_button_cancel)
    dialog_wait $w(modal_packinfo) 0

    select_package_name $package_info(name)
    return $dialog_retval
}

# this is the dialog that shows information
# This is a nicer replacement for the ugly default tk_dialog or tk_message
# for simple Ok-response dialogs

proc dialog_modal_info { window_title bitmap title text } {
    global dialog_retval
    global w
    set dialog_retval 0

    set w(modal_info) [toplevel .modal_info]
    wm title $w(modal_info) $window_title

    set w(info_frame_info) [frame $w(modal_info).frame_info -padx 10 -pady 10]
    
    set w(info_label_title) [label $w(info_frame_info).label_title -text "$title" -font font_subtitle -foreground "darkred" -justify center]

    set w(info_label_icon) [label $w(info_frame_info).label_icon -bitmap $bitmap -foreground "midnightblue" -justify left]

    set w(info_text) [text $w(info_frame_info).info_text -padx 25 -pady 12 -relief sunken -bd 1 -width 60]

    set w(info_scroll) [scrollbar $w(modal_info).info_scroll -command "$w(info_text) yview" ]

    $w(info_text) configure -yscrollcommand "$w(info_scroll) set"
    $w(info_text) insert end "$text\n"

    set w(info_frame_actions) [frame $w(modal_info).frame_actions -padx 10 -pady 10]

    set w(info_button_ok) [button $w(info_frame_actions).button_ok -width 5 -text "Ok" -command { 
	destroy $w(modal_info)
    }]


    pack $w(info_scroll) -side right -fill y
    pack $w(info_frame_info) -side top
    pack $w(info_label_title) -side top -fill x
    pack $w(info_label_icon) -side left -padx 10

    $w(info_text) configure -height 20
    $w(info_text) configure -state disabled

    pack $w(info_text) -side top -padx 10
    
    pack $w(info_frame_actions) -side top
    pack $w(info_button_ok) -side top

    focus $w(info_button_ok)
    dialog_wait $w(modal_info) 0
    
    return $dialog_retval
}

# this one is much simpler, for small messages

proc dialog_modal_msg { window_title bitmap title text } {
    src_dialog $window_title $text $bitmap 0 "Ok"
}


# this dialog gets a username and a password to perform operations as
# another user. Result is a list of two elements, or empty string on cancel

proc dialog_modal_login {} {
    global dialog_retval
    global w

    # this is the return value
    set dialog_retval ""

    set w(modal_login) [toplevel .modal_login]
    wm title $w(modal_login) "Login"

    set w(login_frame_user) [frame $w(modal_login).frame_user -padx 10 -pady 10]
    set w(login_label_user) [label $w(login_frame_user).label_user -text "User"]
    set w(login_entry_user) [entry $w(login_frame_user).entry_user -width 15]
    $w(login_entry_user) insert 0 "root"

    set w(login_frame_pass) [frame $w(modal_login).frame_pass -padx 10 -pady 10]
    set w(login_label_pass) [label $w(login_frame_pass).label_pass -text "Pass"]
    set w(login_entry_pass) [entry $w(login_frame_pass).entry_pass -width 15 -show "*"]
    
    set w(login_frame_actions) [frame $w(modal_login).frame_actions -padx 10 -pady 10]

    set w(login_button_ok) [button $w(login_frame_actions).button_ok -text "Ok" -width 5 -command { 
	set dialog_retval [list [$w(login_entry_user) get] [$w(login_entry_pass) get]]
	destroy $w(modal_login)
    }]

    set w(login_button_cancel) [button $w(login_frame_actions).button_cancel -text "Cancel" -width 5 -command { 
	set dialog_retval ""; destroy $w(modal_login)
    }]
    
    pack $w(login_frame_user) -side top -fill x
    pack $w(login_label_user) -side left
    pack $w(login_entry_user) -side left

    pack $w(login_frame_pass) -side top -fill x
    pack $w(login_label_pass) -side left
    pack $w(login_entry_pass) -side left

    pack $w(login_frame_actions) -side top
    pack $w(login_button_ok) -side left -padx 8
    pack $w(login_button_cancel) -side left -padx 8

    focus $w(login_button_ok)
    dialog_wait $w(modal_login) 0
    
    return $dialog_retval
}

# loads global array package_info
proc registry_load_package { package_name } {
    global packdir
    global package_info
    global w
    set keywords {}

    set title "Registry package load"

    array set package_info [new_package_info]

    if { [catch { open "$packdir/$package_name" "r" } file_packinfo] } {
	src_error $title "Could not open package $package_name.\n\n$file_packinfo"
	registry_load_list
	return 1
    }

    if { [catch { read $file_packinfo } data] } {
	close $file_packinfo
	src_error $title "Could not read package data for $package_name.\n\n$data"
	registry_load_list
	return 1
    }

    close $file_packinfo
    set lines [split $data "\n"]

    foreach line $lines {
	foreach keyword { name description d configured source_available source_size source_location build_subdir installed installed_size install_prefix i } {
	    set pattern "^${keyword}: +(.+)"
	    
	    if { [regexp -nocase $pattern $line matchvar submatch] } {

		if { $keyword == "d" } {
		    if { $package_info(long_description) != "" } {
			append package_info(long_description) "\n"
		    }
		    append package_info(long_description) $submatch

		} elseif { $keyword == "i" } {
		    lappend package_info(installed_files) $submatch

		} else {
		    set package_info($keyword) $submatch
		}
		break
	    }
	}
    }

    # validate package information

    if { $package_info(source_available) == "yes" } {
	if { $package_info(source_location) == "" } {
	    src_warning $title "Source location information missing!"

	} elseif { ![file exists $package_info(source_location)] } {
	    src_warning $title "Source package marked as available is missing:\n\n$package_info(source_location)"

	}
    }

    return 0
}

# select package PACKAGE_INDEX from the list and show relative information

proc select_package_index { package_index } {
    global packages
    global package_info
    global w

    . configure -cursor watch
    update idletasks

    $w(list_packages) selection clear 0 end
    $w(list_packages) selection set $package_index
    set selected_package [lindex $packages $package_index]

    if { [registry_load_package $selected_package] == 0 } {
	set package_info(selected) yes
	show_selected_package
	update_actions
    } else {
	set package_info [new_package_info]
    }

    . configure -cursor left_ptr
}

# select a package from its PACKAGE_NAME

proc select_package_name { package_name } {
    global packages

    set package_index [lsearch -exact $packages $package_name]

    src_assert {$package_index != -1} [expr $package_index != -1]

    select_package_index $package_index
}

proc show_selected_package {} {
    global package_info
    global w

    if { $package_info(selected) == "no" } {
	# package info must be loaded before calling this function
	return 1;
    }

    $w(menu_edit) entryconfigure 4 -state normal

    console_clear
    console_print_title "Package Information "
    console_print_ref edit_info {[E]dit} {
	dialog_modal_package_info
    }
    console_print "\n\n"

    console_print_subtitle "package name: "
    console_print "$package_info(name)\n"
    
    if { $package_info(configured) != "no" } {
	console_print_subtitle "configured: "
	console_print "$package_info(configured)\n"
    }

    if { $package_info(build_subdir) != "" } {
	console_print_subtitle "build subdirectory: "
	console_print "$package_info(build_subdir)\n"
    }

    if { $package_info(source_available) != "" } {
	console_print_subtitle "source available: "
	console_print "$package_info(source_available)\n"
    }

    if { $package_info(source_location) != "" } {
	console_print_subtitle "source location: "
	console_print "$package_info(source_location)\n"
    }

    if { $package_info(source_size) != "" } {
	console_print_subtitle "compressed source size (KB): "
	console_print "$package_info(source_size)\n"
    }

    if { $package_info(installed) != "" } {
	console_print_subtitle "installed: "
	console_print "$package_info(installed)\n"
    }

    if { $package_info(installed) == "yes" } {
	if { $package_info(installed_size) != "" } {
	    console_print_subtitle "installed size (KB): "
	    console_print "$package_info(installed_size)\n"
	}

	# install_check value is 0 if ok, 1 if some error
	# installed_files_exist is a list of 0=ok, 1=err for each filename
	set len [llength $package_info(installed_files)]
	if { $len > 0 } {
	    set install_check 0
	    set installed_files_exist [list]
	    for { set i 0 } { $i < $len } { incr i } {
		set filename [lindex $package_info(installed_files) $i]
		if { [file_ltype $filename] != "none" } {
		    lappend installed_files_exist 0
		} else {
		    lappend installed_files_exist 1
		    set install_check 1
		}
	    }

	    if { $install_check } {
		console_print_err "\nSome files written during installation seem deleted or unreachable (***)\n"
	    } 
	}
	
	console_print_subtitle "\ninstalled files and directories:\n"

	if { $len > 0 } {
	    for { set i 0 } { $i < $len } { incr i } {
		set filename [lindex $package_info(installed_files) $i]
		if { [lindex $installed_files_exist $i] == 0 } {
		    console_print "$filename\n"
		} else {
		    console_print_err "$filename (***)\n"
		}
	    }
	} else {
	    console_print_err "no information about installed files available."
	}
    }

    if { $package_info(description) != "" } {
	console_print_subtitle "\npackage description: "
	console_print "$package_info(description)\n\n"
	console_print "$package_info(long_description)\n\n"
    }

    console_home

    return 0
}

# remove and recreate build directories
proc clear_builddirs {} {
    global builddir
    global destdir
    set title "Clear build directories"

    if { [catch { file delete -force $builddir } errstr] } {
	src_warning $title "Could not clear build directory.\n\n$errstr"
	return 1
    }

    if { [catch { file delete -force $destdir } errstr] } {
	src_warning $title "Could not clear temporary install directory.\n\n$errstr"
	return 1
    }

    # recreate them
    src_directories_check
}

proc registry_remove_package { local_package_info_src } {
    global packdir

    array set local_package_info $local_package_info_src
    set title "Registry remove package"

    if { [file_type "$packdir/$local_package_info(name)"] != "file" } {
	return 1
    }

    if { [catch { file delete "$packdir/$local_package_info(name)" } errstr] } {
	src_error $title "Could not delete package information.\n\n$errstr"
	return 1
    }

    registry_load_list
    return 0
}

proc registry_add_package { local_package_info_src } {
    global packdir

    array set local_package_info $local_package_info_src
    set title "Registry add package"

    if { [file_type "$packdir/$local_package_info(name)"] != "none" } {
	console_print_comment "overwriting previous package information"
    }
    
    if { [catch { open "$packdir/$local_package_info(name)" "w" } fh] } {
	src_error $title "Could not open package $package_name.\n\n$fh"
	return 1
    }

    set data ""

    foreach keyword [array names local_package_info] {
	if { $local_package_info($keyword) != "" && $keyword != "selected" } {
	    set line [string toupper $keyword]
	    append line ": "

	    if { $keyword == "installed_files" } {
		append line "\n"
		foreach var $local_package_info($keyword) {
		    if { $var != "" } {
			append line "i: $var\n"
		    }
		}

	    } elseif { $keyword == "long_description" } {
		append line "\n"
		foreach var [split $local_package_info($keyword) "\n"] {
		    append line "d: $var\n"
		}

	    } else {
		append line "$local_package_info($keyword)"
	    }

	    append data "$line\n"
	}
    }

    if { [catch { puts -nonewline $fh $data } errstring] } {
	close $fh
	src_error $title "Could not write package data.\n\n$errstring"
	return 1
    }

    close $fh
    registry_load_list
    return 0
}

# return a list of all file and link dependencies relative to current package in
# package_info

proc get_relative_dependencies {} {
    global packages
    global package_info

    set retval [list]
    set installed_files $package_info(installed_files)

    for { set i 0 } { $i < [llength $installed_files] } { incr i } {
	set filename [lindex $installed_files $i]
	set t [file_ltype $filename]

	if { $t != "file" && $t != "link" } {
	    # check only files and links for dependencies. Dirs will be
	    # handled later by caller
	    set installed_files [lreplace $installed_files $i $i]
	    incr i -1
	}
    }

    array set saved_info [array get package_info]

    foreach this $packages {
	if { $saved_info(name) == $this } { continue }
	if { [registry_load_package $this] == 0 } {
	    if { [llength $package_info(installed_files)] == 0 } {
		continue
	    }
	    foreach this_file $package_info(installed_files) {
		if { [lsearch -exact $installed_files $this_file] != -1 } {
		    lappend retval $this_file
		}
	    }
	}
    }

    array unset package_info
    array set package_info [array get saved_info]
    return $retval
}

# status bar handling

proc show_status_bar { todo_list } {
    global w
    
    set argc [llength $todo_list]

    set width [expr [winfo width $w(frame_status)]/$argc]

    $w(frame_status) configure -height 40 -background antiquewhite

    set w(label_status_command) [label $w(frame_status).command -text "Running:" -foreground white -background black -font font_subtitle -wraplength [expr [winfo width $w(frame_status)] - 50]]

    pack $w(label_status_command) -fill x -side top

    for { set i 0 } { $i < $argc } { incr i } {
	set t [lindex $todo_list $i]
	set w("frame_status_$t") [frame "$w(frame_status).frame_status_$t" -background black -width $width]
	set w("label_status_$t") [label $w("frame_status_$t").label_status_$t -background black -text "$t" -foreground white -font font_text]

	pack $w("frame_status_$t") -side left -expand 1 -fill both
	pack $w("label_status_$t") -expand 1 -fill both
    }

}

proc hide_status_bar {} {
    global w

    foreach this [winfo children $w(frame_status)] {
	destroy $this
    }

    $w(frame_status) configure -height 1 -background antiquewhite3
}

# a bit too much direct, will probably change
proc set_status_bar_item { t value } {
    global w
    
    if { $value == "yellow" } {
	$w("label_status_$t") configure -background cyan -foreground black -font font_subtitle

    } elseif { $value == "green" } {
	$w("label_status_$t") configure -background blue -foreground white -font font_subtitle
    } elseif { $value == "red" } {
	$w("label_status_$t") configure -background red -foreground white -font font_subtitle
    }

    #    $w("label_status_$t") configure -background $value
}

proc status_print_command { text } {
    global w
    
    if { [lsearch -exact [array names w] label_status_command] != -1 && [winfo exists $w(label_status_command)] } {
	$w(label_status_command) configure -text "Running: $text"
    }
}

proc action_add {} {
    global dialog_retarr

    set retval 0

    lock_actions
    update_actions

    dialog_modal_add
    set filename $dialog_retarr(filename)
    set packagename $dialog_retarr(packagename)

    if { $filename == "" } {
	# do nothing

    } else {
	set retval [add_package $filename $packagename]
    }

    unlock_actions
    update_actions

    return $retval;
}

proc action_install {} {
    global spawn_id
    global package_info
    global o
    global sigint_seen

    set title "Install action"

    if { $spawn_id != 0 } {
	# install button is currently showing the "stop"
	catch { exec /bin/kill -s INT [exp_pid] }
	incr sigint_seen
	return 0
    }

    # assert we have the source
    src_assert {$package_info(source_available) == "yes"} [expr $package_info(source_available) == "yes"]

    lock_actions
    update_actions

    console_clear

    set o(install) 1
    set o(keep_source) 1

    if { $package_info(installed) == "yes" } {
	console_print_title "reinstall $package_info(name)\n\n"

	if { [uninstall_package] } {
	    unlock_actions
	    update_actions
	    return 1
	}

    } else {
	console_print_title "install $package_info(name)\n\n"
    }

    set reconf [src_dialog "Reconfigure package" \
		    "Do you want to reconfigure the package manually?\n" \
		    question \
		    0 \
		    "No" "Yes"]
    
    if { $reconf < 0 } { set reconf 0 }

    set result [install_package $package_info(source_location) $reconf]
    clear_builddirs

    if { $result == 0 } {
	registry_add_package [array get package_info]
	src_message $title "Package successfully installed."
    } else {
	src_warning $title "Package has NOT been installed."
    }

    unlock_actions
    update_actions

    return $result
}

proc uninstall_package {} {
    global package_info
    global sigint_seen

    set title "Uninstall package"
    set int_message "Uninstallation suspended by user. Check the package for errors by clicking the package name in the list."

    set result 1
    console_print_comment "uninstall package"

    if { $package_info(source_available) == "yes" } {
	lappend todo_list "Unpack source" "Uninstall"
    }

    lappend todo_list "Internal uninstall" "Empty directories"
    
    show_status_bar $todo_list

    if { $package_info(source_available) == "yes" } {
	set result [make_uninstall]
	clear_builddirs
	if { $result != 0 && $sigint_seen != 0 } {
	    # if STOP is requested by user, suspend uninstallation immediately.
	    src_message $title $int_message
	    hide_status_bar
	    return 1
	}
    }

    set result [internal_uninstall]
    if { $result != 0 && $sigint_seen != 0 } {
	# if STOP is requested by user, suspend uninstallation immediately.
	# At this point, if the package is conforming, it's probably far too
	# late. However, errors will be shown and one can use 'Reinstall' 
	# to fix things up.

	src_message $title $int_message
	hide_status_bar
	return 1
    }
    
    if { $result == 0 } {
	set package_info(installed) "no"
    }

    hide_status_bar
    return $result
}

proc action_remove {} {
    global package_info
    set title "Removal action"

    lock_actions
    update_actions

    set what [dialog_modal_remove]
    # 0=removed 1=not removed 
    set removed_bin 1
    set removed_src 1
    # if skip src is 1 then skip source removal
    set skip_src 0

    if { $what > 0 } {
	console_clear
	console_print_title "remove $package_info(name)\n\n"
    }

    if { [expr $what & 1] } {
	set removed_bin [uninstall_package]
	if { $removed_bin != 0 } {
	    # if removal of binaries failed, do NOT proceed with source
	    # uninstallation.
	    set skip_src 1
	}

    } elseif { $what == 2 && $package_info(installed) == "yes" && [llength $package_info(installed_files)] == 0 } {
	# if removing only the source and no info about installed files is
	# available
	set skip_src [src_dialog "Source removal" \
			  "Removing source code from this non-conforming package will make it impossible to uninstall it automatically. Really proceed with the source removal?" question 1 "Yes" "No"]
	if { $skip_src < 0 } { set skip_src 1 }
    }
    
    if { [expr $what & 2] && $skip_src != 1 } {
	# remove source
	
	console_print_comment "remove package source"
	set removed_src [source_uninstall]
	if { $removed_src == 0 } {
	    set package_info(source_available) "no"
	    set package_info(source_location) ""
	    set package_info(build_subdir) ""
	}
    }
    
    if { $removed_bin == 0 || $removed_src == 0 } {
	if { $package_info(source_available) == "yes" || \
		 $package_info(installed) == "yes" } {
	    registry_add_package [array get package_info]
	    
	} else {
	    registry_remove_package [array get package_info]
	    array set package_info [ new_package_info ]
	}

	src_message $title "Removal operation successful."

    } else {
	# nothing to do
    }
    
    unlock_actions
    update_actions
}

proc update_actions {} {
    global package_info
    global w
    global actions_locked
    global spawn_id

    if { $actions_locked } {
	$w(button_remove) configure -state disabled
	$w(button_add) configure -state disabled
	$w(list_packages) configure -state disabled
	$w(menu_edit) entryconfigure 4 -state disabled
	$w(menu_edit) entryconfigure 6 -state disabled

	if { $spawn_id != 0 } {
	    . configure -cursor watch
	    $w(button_install) configure -state normal
	    $w(button_install) configure -text "Stop"
	    focus $w(button_install)

	} else {
	    $w(button_install) configure -state disabled
	}
	return 0
    }

    . configure -cursor left_ptr

    $w(list_packages) configure -state normal
    $w(menu_edit) entryconfigure 6 -state normal

    if { $package_info(selected) == "yes" } {
	# a package is selected
	$w(button_remove) configure -state normal

	if { $package_info(source_available) != "yes" } {
	    $w(button_install) configure -state disabled

	} else {
	    $w(button_install) configure -state normal

	    if { $package_info(installed) == "yes" } {
		$w(button_install) configure -text "Reinstall" -underline 2
		# focus $w(button_remove)
	    } else {
		$w(button_install) configure -text "Install" -underline 0
		# focus $w(button_install)
	    }
	}

	$w(button_add) configure -state normal
	$w(menu_edit) entryconfigure 4 -state normal

    } else {
	# no packages selected
	$w(button_remove) configure -state disabled
	$w(button_install) configure -state disabled
	$w(button_add) configure -state normal
	focus $w(button_add)
	$w(button_install) configure -text "Install" -underline 0
	$w(menu_edit) entryconfigure 4 -state disabled
    }
    return 0
}

proc show_popup_action { abs_x abs_y rel_x rel_y } {
    global w
    global packages
    global package_info
    global actions_locked
    global spawn_id

    if { !$actions_locked && $spawn_id == 0 && [llength $packages] > 0 } {
	select_package_index [$w(list_packages) nearest $rel_y]

	if { $package_info(source_available) != "yes" } {
	    $w(popup_action) entryconfigure 1 -state disabled

	} else {
	    $w(popup_action) entryconfigure 1 -state normal

	    if { $package_info(installed) == "yes" } {
		$w(popup_action) entryconfigure 1 -label "Reinstall" -underline 2
	    } else {
		$w(popup_action) entryconfigure 1 -label "Install" -underline 0
	    }
	}

	tk_popup $w(popup_action) $abs_x $abs_y
	return 0
    }
    return 1
}

# initial and global event bindings

proc bind_events {} {
    global packages
    global w
    global package_info
    global actions_locked
    global spawn_id

    bind $w(list_packages) <Button-2> {
	show_popup_action %X %Y %x %y
    }

    bind $w(list_packages) <Button-3> {
	show_popup_action %X %Y %x %y
    }

    bind $w(list_packages) <<ListboxSelect>> {
	if { !$actions_locked && $spawn_id == 0 } {
	    if { [%W curselection] != "" } {
		select_package_index [%W curselection]
	    }
	}
    }

    bind . <KeyPress-Return> {
	event generate . <KeyPress-space>
    }

    bind . <KeyRelease-Return> {
	event generate . <KeyRelease-space>
    }

    bind . a {
	$w(button_add) invoke
    }

    bind . i {
	$w(button_install) invoke
    }
    
    bind . r {
	$w(button_remove) invoke
    }

    bind . e {
	if { $package_info(selected) == "yes" } {
	    dialog_modal_package_info
	}
    }

    trap handle_sigint { SIGINT }
    trap clean_exit(2) { SIGTERM }
}

proc handle_sigint {} {
    global spawn_id
    global sigint_seen

    incr sigint_seen
    
    if { $spawn_id != 0 && $sigint_seen < 2 } {
	# install button is currently showing the "stop"
	catch { exec /bin/kill -s INT [exp_pid] }

    } else {
	clean_exit 2
    }
}

proc parse_cmdline { argc argv } {
    foreach option $argv {
	if { $option == "-h" || $option == "--help" } {
	    srcinst_help
	    exit 0
	    
	} elseif { $option == "-V" || $option == "--version" } {
	    srcinst_version
	    exit 0
	}
    }
}

proc srcinst_help {} {
    global env

    puts "\nUsage: sourceinstall [-h] [-V] [SOURCE_PACKAGE, ...]\n"
    puts "If called without any arguments, the program starts a graphical interface that can be used to install, remove and query information about installed packages.\n"
    puts "Options:\n"
    puts "-h, --help      Display this help and exit successfully"
    puts "-V, --version   Display a version string and exit successfully\n"
    puts "If one or more SOURCE_PACKAGE arguments are specified, they are assumed to be referring to valid source packages to add (honouring the current preferences about package adding)."
    puts "Packages are added in the same order as specified on the command line, but the first failure in the installations breaks the chain."
    puts "You can change the preferences using the graphical interface under Edit->Preferences, or you can edit the configuration file\n"
    puts "$env(HOME)/.sourceinstall/sourceinstallrc\n"
    puts "Type info sourceinstall or man sourceinstall for more information."
    puts "Report bugs to bug-sourceinstall@gnu.org"
}

proc srcinst_version {} {
    global env

    puts "GNU Source Installer (sourceinstall) version 0.4,"
    puts "Copyright (c) 2005 Claudio Fontana"
    puts "This is Free Software that comes with NO WARRANTY;"
    puts "(see README.license for more details)"
}

proc srcinst { argc argv } {

    if { $argc > 0 } {
	lock_actions
	update_actions

	foreach filename $argv {
	    set fullname [file normalize $filename]
	    if {[add_package $fullname ""]} { break }
	}
	
	unlock_actions
	update_actions
    }
}

proc registry_load_list {} {
    global packages;
    global packdir;

    lset packages [lsort [glob -directory $packdir -type f -tails -nocomplain * ] ]
}

proc options_save {} {
    global cfgfile
    global o

    set title "Options save"

    if { [catch { open $cfgfile "w" } fh] } {
	src_warning $title "Could not create configuration file.\n\n$fh"
	return 1
    }

    foreach keyword [ array names o ] {
	if { $o($keyword) == "" } { continue }
	set line [string toupper $keyword]
	append line ": $o($keyword)"
	if { [catch { puts $fh $line } err] } {
	    close $fh
	    src_warning $title "Could not write configuration file.\n\n$err"
	    return 1
	}
    }

    close $fh
    return 0
}

proc options_load {} {
    global cfgfile
    global o
    set title "Options load"

    if { ![file exists $cfgfile] } {
	return [options_save]
    }

    if { [catch { open $cfgfile "r" } fh] } {
	src_warning $title "Could not open configuration file,\nreverting to defaults.\n\n$fh"
	return 1
    }

    if { [catch { read $fh } data] } {
	close $fh
	src_warning $title "Could not read configuration file,\nreverting to defaults.\n\n$fh"
	return 1
    }
    
    set lines [split $data "\n"]
    
    foreach line $lines {
	foreach keyword [ array names o ] {
	    set pattern "^${keyword}: *(.+)"
	    if { [regexp -nocase $pattern $line matchvar submatch] } {
		set o($keyword) $submatch
	    }
	}
    }
    close $fh
    return 0
}

# fonts definitions
proc init_fonts {} {
    font create font_button -family Helvetica -size 10 -weight normal
    font create font_title -family Helvetica -size 14 -weight normal
    font create font_subtitle -family Helvetica -size 11 -weight bold
    font create font_text -family Helvetica -size 11 -weight normal
    font create font_small -family Helvetica -size 8 -weight normal
    font create font_ref -family Helvetica -size 11 -underline 1

    font create font_code -family Courier -size 11 -weight normal
    font create font_command -family Courier -size 11 -weight bold -slant italic
    font create font_comment -family Courier -size 11 -weight normal -slant italic
}

# default options for widgets
proc init_widget_defaults {} {
    option add *font font_text
    option add *Label.font font_text
    option add *Button.font font_button

    # remove highlightthickness from widgets
    # option add *highlightThickness 0
    # enable that for buttons
    option add *Button.highlightThickness 1

    # widgets default colors
    option add *background antiquewhite3
    
    option add *Text.background antiquewhite
    option add *Text.foreground midnightblue
    option add *Entry.background antiquewhite
    option add *Entry.foreground midnightblue
    option add *Scrollbar.background antiquewhite
    option add *Listbox.background antiquewhite
    option add *Listbox.foreground midnightblue
    option add *Button.background antiquewhite2
    
    option add *selectBackground \#ffffa0
    option add *activeBackground \#ffffc0
    option add *troughColor antiquewhite2
    option add *highlightColor midnightblue
    option add *highlightBackground antiquewhite3
    option add *insertBackground midnightblue

    # other widgets defaults

    option add *Menu.tearOff 0
    option add *Menu.relief raised
    option add *Menu.borderWidth 1

    option add *Button.relief raised
    option add *Button.relief raised
    option add *Button.relief raised

    option add *Text.wrap word
    option add *Scrollbar.elementBorderWidth 2
    option add *Scrollbar.width 10
    option add *Labelframe.font font_text
    option add *Labelframe.foreground darkred
    option add *Label.foreground darkred
}

proc src_directories_check {} {
    global cfgdir
    global packdir
    global srcdir
    global builddir
    global destdir

    set title "Program directories check"

    set directories [list $cfgdir $packdir $srcdir $builddir $destdir]

    foreach var $directories {
	set t [file_type $var]
	if { $t != "none" } {
	    if { [ expr { $t != "directory" || ![file readable $var] || ![file writable $var] } ] } {
		src_error $title "Cannot gain read/write access to directory $var\n\n"
		clean_exit 1
	    }

	} else {
	    if { [catch { file mkdir $var } fh] } {
		src_error $title "Cannot create directory $var\n\n$fh"
		clean_exit 1
	    }
	}
    }
}

proc sourceinstall_lock {} {
    global lockfile
    global env

    if { [catch { open $lockfile [list RDWR CREAT EXCL] } fh] } {
	puts stderr "$fh\n"
	if { [file_type $lockfile] != "none" } {
	    if { [catch { open $lockfile "r" } fh] ||
		 [catch { read -nonewline $fh } pid] } {
		puts stderr "lock seems to exist, but it is unreadable.\n$fh"
		exit 1
	    }
	    close $fh

	    set running 0
	    set timeout -1
	    if { [catch { eval spawn -noecho "/bin/ps -p $pid" }] } {
		set running -1

	    } else {
		expect {
		    eof {
		    }
		    
		    -re ".*\n" {
			if { [regexp $pid $expect_out(buffer)] } {
			    incr running
			}
			
			if { [regexp {sourceinstall} $expect_out(buffer)] } {
			    incr running
			}
			exp_continue
		    }
		}
		wait
	    }

	    if { $running > 1} {
		puts stderr "\nYou are almost certainly running another instance of sourceinstall as the same user. The processes list confirms that. It is recommended to have only one running for each user."
	    } elseif { $running == 1 } {
		puts stderr "\nYou appear to be running another instance of sourceinstall as the same user. However the processes list does not completely confirm that (sourceinstall could not be found as process name)."
	    } elseif { $running == 0 } {
		puts stderr "\nsourceinstall reports another running instance as the same user, but it could not be found in the processes list."
	    } elseif { $running == -1 } {
		puts stderr "\nsourceinstall reports another running instance as the same user, but the 'ps' command is not available, so it can not be confirmed by a process list."
	    }
	    
	    puts stderr "\nIf you are sure that no other instance of sourceinstall is running as the same user, remove $lockfile yourself."
	}
	exit 1
    }

    if { [catch { puts $fh [pid] }] } {
	puts stderr "Warning: could not write lock file $lockfile"
    }

    close $fh
}

proc clean_exit { code } {
    global lockfile

    if { [file_type $lockfile] == "none" } {
	puts stderr "Warning: lockfile does not exist: $lockfile"

    } elseif { [catch { file delete $lockfile }] } {
	puts stderr "Warning: could not remove lockfile: $lockfile"
    }
    exit $code
}

# startup

parse_cmdline $argc $argv

catch { package require Tk }
catch { package require Expect }
set spawn_id 0 ; set timeout -1; log_user 0

src_directories_check
sourceinstall_lock

init_fonts
init_widget_defaults
init_wm

options_load
registry_load_list
bind_events
update_actions

set about "Copyright (c) 2005 Claudio Fontana

GNU Source Installer comes with ABSOLUTELY NO WARRANTY; this is free software, and you are welcome to redistribute it under certain conditions. See the README.license file that came with the package for full details. You can contact the original author by sending an email to claudio@gnu.org.

You can reach the GNU Source Installer home page at
http://www.gnu.org/software/sourceinstall

Send your help requests, feature wishes, general feedback and bug reports to
bug-sourceinstall@gnu.org

Development takes place at the savannah project page at
http://savannah.gnu.org/projects/sourceinstall

This program is part of the GNU project. You can learn more about the GNU project by browsing
http://www.gnu.org
"
console_intro
srcinst $argc $argv

